"""Data generation for linear model unit tests."""
import numpy as np
from brainspace.mesh.mesh_elements import get_points
from sklearn.model_selection import ParameterGrid

from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect
from brainstat.tests.testutil import (
    _generate_sphere,
    array2effect,
    generate_random_data_model,
    save_input_dict,
    slm2files,
)


def generate_test_data():
    np.random.seed(0)
    surface = _generate_sphere()
    parameters = [
        {
            "n_observations": [103],
            "n_vertices": [np.array(get_points(surface)).shape[0]],
            "n_variates": [1, 2, 3],
            "n_predictors": [1, 7],
            "n_random": [0],
            "surf": [None, surface],
        },
        {
            "n_observations": [103],
            "n_vertices": [np.array(get_points(surface)).shape[0]],
            "n_variates": [1],
            "n_predictors": [2, 7],
            "n_random": [1],
            "surf": [None, surface],
        },
    ]

    test_num = 0
    for params in ParameterGrid(parameters):
        test_num += 1
        Y, M = generate_random_data_model(
            params["n_observations"],
            params["n_vertices"],
            params["n_variates"],
            params["n_predictors"],
        )

        save_input_dict(
            {"Y": Y, "M": M, "surf": params["surf"], "n_random": params["n_random"]},
            "xlinmod",
            test_num,
        )

        model = array2effect(M, params["n_random"])

        slm = SLM(model, FixedEffect(1), params["surf"])
        slm.linear_model(Y)
        slm2files(slm, "xlinmod", test_num)


if __name__ == "__main__":
    generate_test_data()



from pathlib import Path
from typing import Callable, Optional, Union

import numpy as np
from brainspace.utils.parcellation import reduce_by_labels

from brainstat._typing import ArrayLike
from brainstat.datasets import fetch_gradients, fetch_parcellation


def yeo_networks_associations(
    data: ArrayLike,
    template: str = "fsaverage5",
    seven_networks: bool = True,
    data_dir: Optional[Union[str, Path]] = None,
    reduction_operation: Union[str, Callable] = np.nanmean,
) -> np.ndarray:
    """Computes association

    Parameters
    ----------
    data : ArrayLike
        Data to be summarized in the Yeo networks in a sample-by-feature format.
    template : str, optional
        Surface template. Valid values are "fsaverage5", "fsaverage", and
        "fslr32k", by default "fsaverage5".
    seven_networks : bool, optional
        If true, uses the 7 network parcellation, otherwise uses the 17 network
        parcellation, by default True.
    data_dir : str, Path, optional
        Data directory to store the Yeo network files, by default $HOME_DIR/brainstat_data/parcellation_data.
    reduction_operation : str, callable, optional
        How to summarize data. If str, options are: {‘min’, ‘max’, ‘sum’,
        ‘mean’, ‘median’, ‘mode’, ‘average’}. If callable, it should receive a
        1D array of values, array of weights (or None) and return a scalar
        value. Default is ‘mean’.

    Returns
    -------
    np.ndarray
        Summary statistic in the yeo networks.
    """
    n_regions = 7 if seven_networks else 17
    yeo_networks = fetch_parcellation(
        template=template,
        atlas="yeo",
        n_regions=n_regions,
        join=True,
        data_dir=data_dir,
    )

    if np.array(data).ndim == 1:
        data_2d = np.array(data)[:, None]
    else:
        data_2d = np.array(data)

    n_features = data_2d.shape[1]

    yeo_mean = np.zeros((n_regions + 1, n_features))
    for i in range(n_features):
        yeo_mean[:, i] = reduce_by_labels(
            data_2d[:, i], yeo_networks, red_op=reduction_operation
        )
    return yeo_mean[1:, :]


def gradients_corr(
    data: ArrayLike,
    name: str = "margulies2016",
    template: str = "fsaverage5",
    data_dir: Optional[Union[str, Path]] = None,
    overwrite: bool = False,
) -> np.ndarray:
    """Comptues the correlation of the input data with the Margulies gradients.

    Parameters
    ----------
    data : ArrayLike
        The data to be compared to the Margulies gradients. Data must be in the
        shape of vertices-by-features.
    name : str, optional
        Name of the gradients. Valid values are "margulies2016", defaults to
        "margulies2016".
    template : str, optional
        Name of the template surface. Valid values are "fsaverage5",
        "fsaverage7", "fslr32k", defaults to "fsaverage5".
    data_dir : str, Path, optional
        Path to the directory to store the Margulies gradient data files, by
        default $HOME_DIR/brainstat_data/functional_data.
    overwrite : bool, optional
        If true, overwrites existing files, by default False.

    Returns
    -------
    np.ndarray
        Correlations between the input data and the Margulies gradients.
    """
    gradients = fetch_gradients(
        name=name, template=template, data_dir=data_dir, overwrite=overwrite
    )
    return _columnwise_correlate(data, gradients)


def _columnwise_correlate(x: ArrayLike, y: Optional[ArrayLike] = None) -> np.ndarray:
    """Implements MATLAB's corr function for Pearson correlations of each column
    in x with each column in y. If y is not provided, computes correlation with
    x onto itself.

    Parameters
    ----------
    x : ArrayLike
        2D data matrix.
    y : ArrayLike, optional
        2D data matrix.

    Implements the function R = (1/n-1) @ X_s' @ Y_s.
    - n is the number of samples.
    - X_s and Y_s are standardized versions of X and Y i.e. X_s = C @ X @ D.
    - C is the centering matrix.
    - D is a scaling matrix i.e. a diagonal matrix with std(x, axis=0) on the
        diagonal.

    Returns
    -------
    np.ndarray
        Pearson correlation matrix.
    """

    n_samples, n_features = np.shape(x)
    centering = np.identity(n_samples) - 1 / n_samples
    scaling = lambda v: np.identity(n_features) * np.std(v, axis=0, ddof=1)
    centered_scaled = lambda v: centering @ v @ np.linalg.inv(scaling(v))

    if y is None:
        X_s = centered_scaled(x)
        return 1 / (n_samples - 1) * X_s.T @ X_s
    else:
        return 1 / (n_samples - 1) * centered_scaled(x).T @ centered_scaled(y)



"""Data generation for compute_resels unit tests."""

import pickle

import numpy as np
from brainspace.mesh.mesh_elements import get_cells, get_points
from nilearn import datasets

from brainstat.context.utils import read_surface_gz
from brainstat.stats._multiple_comparisons import compute_resels
from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect
from brainstat.tests.testutil import datadir


def generate_random_slm(rand_dict):
    """Generates a valid SLM for a surface.
    Parameters
    ----------
    surf : BSPolyData or a dictionary with key 'tri'
        Brain surface.
    Returns
    -------
    brainstat.stats.SLM
        SLM object.
    """
    # this is going to be the input slm
    I = {}
    rand_slm = SLM(FixedEffect(1), FixedEffect(1))
    for key in rand_dict.keys():
        setattr(rand_slm, key, rand_dict[key])
        I[key] = rand_dict[key]

    # this is going to be the output dict
    O = {}
    O["resels"], O["reselspvert"], O["edg"] = compute_resels(rand_slm)

    return I, O


def params2files(I, O, test_num):
    """Converts params to input/output files"""
    # filenames for the input and outpur dict's
    basename = "xstatresl"
    fin_name = datadir(basename + "_" + f"{test_num:02d}" + "_IN.pkl")
    fout_name = datadir(basename + "_" + f"{test_num:02d}" + "_OUT.pkl")

    # save input and output data in pickle format with filenames above
    with open(fin_name, "wb") as f:
        pickle.dump(I, f, protocol=4)
    with open(fout_name, "wb") as g:
        pickle.dump(O, g, protocol=4)
    return


def generate_tests():
    # Test 01
    # ['tri'] will be a np array, shape (4, 3), int64
    np.random.seed(0)
    rand_dict = {}
    rand_dict["tri"] = np.random.randint(1, int(10), size=(4, 3))
    In, Out = generate_random_slm(rand_dict)
    params2files(In, Out, 1)

    # Test 02
    # ['tri'] :np array, shape (4, 3), int64
    # ['resl'] :np array, shape (8, 6), float64
    np.random.seed(0)
    rand_dict = {}
    n_vertices = 6
    rand_dict["tri"] = np.random.randint(1, n_vertices, size=(4, 3))
    rand_dict["resl"] = np.random.random_sample((8, n_vertices))
    In, Out = generate_random_slm(rand_dict)
    params2files(In, Out, 2)

    # Test 03
    # ['tri'] :np array, shape (4, 3), int64
    # ['resl'] :np array, shape (8, 6), float64
    # ['mask'] :np array, shape (5,), bool
    np.random.seed(0)
    rand_dict = {}
    n_vertices = 6
    rand_dict["tri"] = np.random.randint(1, n_vertices, size=(4, 3))
    rand_dict["resl"] = np.random.random_sample((8, n_vertices))
    rand_dict["mask"] = np.random.choice(
        a=[True, False], size=(rand_dict["tri"].max(),)
    )
    In, Out = generate_random_slm(rand_dict)
    params2files(In, Out, 3)

    # Test 04
    # ['lat'] :np array, shape (10, 10, 10), float64
    np.random.seed(0)
    rand_dict = {}
    rand_dict["lat"] = np.ones((10, 10, 10))
    In, Out = generate_random_slm(rand_dict)
    params2files(In, Out, 4)

    # Test 05
    # ['lat'] :np array, shape (10, 10, 10), bool
    np.random.seed(0)
    rand_dict = {}
    rand_dict["lat"] = np.random.choice(a=[False, True], size=(10, 10, 10))
    In, Out = generate_random_slm(rand_dict)
    params2files(In, Out, 5)

    # Test 06
    # ['tri] :np array, shape (1000,3)
    # ['mask'] :np array, shape (['tri'].max(),), bool
    np.random.seed(0)
    rand_dict = {}
    rand_dict["tri"] = np.random.randint(1, n_vertices, size=(1000, 3))
    rand_dict["mask"] = np.random.choice(
        a=[True, False], size=(rand_dict["tri"].max(),)
    )
    In, Out = generate_random_slm(rand_dict)
    params2files(In, Out, 6)

    # Test 07
    # ['lat'] :np array, shape (10, 10, 10), bool
    # ['resl'] :np array, shape (1359, 1), float64
    np.random.seed(0)
    rand_dict = {}
    rand_dict["lat"] = np.random.choice(a=[False, True], size=(10, 10, 10))
    rand_dict["resl"] = np.random.random_sample((1359, 1))
    In, Out = generate_random_slm(rand_dict)
    params2files(In, Out, 7)

    # Test 08
    # ['tri] :np array, shape (1000,3)
    # ['mask'] :np array, shape (499,), bool
    np.random.seed(1)
    rand_dict = {}
    rand_dict["tri"] = np.random.randint(1, 499, size=(1000, 3))
    rand_dict["mask"] = np.random.choice(
        a=[True, False], size=(rand_dict["tri"].max(),)
    )
    In, Out = generate_random_slm(rand_dict)
    params2files(In, Out, 8)

    # Test 09
    # ['lat'] :np array, shape (10, 10, 10), bool
    # ['resl'] :np array, shape (1198, 1), float64
    np.random.seed(1)
    rand_dict = {}
    rand_dict["lat"] = np.random.choice(a=[False, True], size=(10, 10, 10))
    rand_dict["resl"] = np.random.random_sample((1198, 1))
    In, Out = generate_random_slm(rand_dict)
    params2files(In, Out, 9)

    # Test 10
    # ['tri'] is pial_fs5, shape (20480, 3)
    pial_fs5 = datasets.fetch_surf_fsaverage()["pial_left"]
    pial_surf = read_surface_gz(pial_fs5)
    n_vertices = get_points(pial_surf).shape[0]
    rand_dict = {}
    rand_dict["tri"] = np.array(get_cells(pial_surf)) + 1
    In, Out = generate_random_slm(rand_dict)
    params2files(In, Out, 10)

    # Test 11
    # ['tri'] :pial_fs5, shape (20480, 3)
    # ['mask'] :np array, shape (['tri'].max(),), bool
    np.random.seed(0)
    rand_dict = {}
    rand_dict["tri"] = np.array(get_cells(pial_surf)) + 1
    rand_dict["mask"] = np.random.choice(
        a=[True, False], size=(rand_dict["tri"].max(),)
    )
    In, Out = generate_random_slm(rand_dict)
    params2files(In, Out, 11)

    # Test 12
    # ['tri'] :pial_fs5, shape (20480, 3) --> shuffle
    # ['mask'] :np array, shape (['tri'].max(),), bool
    np.random.seed(5)
    rand_dict = {}
    rand_dict["tri"] = np.array(get_cells(pial_surf)) + 1
    np.random.shuffle(rand_dict["tri"])
    rand_dict["mask"] = np.random.choice(
        a=[True, False], size=(rand_dict["tri"].max(),)
    )
    In, Out = generate_random_slm(rand_dict)
    params2files(In, Out, 12)


if __name__ == "__main__":
    generate_tests()



"""Data generation for FDR unit tests."""
import pickle

import numpy as np

from brainstat.stats._multiple_comparisons import fdr
from brainstat.stats._t_test import t_test
from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect
from brainstat.tests.testutil import datadir


def generate_random_fdr_data(
    t_dim,
    df_max,
    finname,
    k=1,
    df_dim=None,
    mask_dim=None,
    dfs_max=None,
    tri_dim=None,
    resl_dim=None,
    du=None,
    c_dim=None,
    ef_dim=None,
    sd_dim=None,
    X_dim=None,
    coef_dim=None,
    SSE_dim=None,
    seed=0,
):
    """Generate random test datasets."""
    # t_dim : tuple
    # df : int
    # k : int
    # finname : filename ending with *pkl
    np.random.seed(seed=seed)
    t = np.random.random_sample(t_dim)

    if df_dim is None:
        df = np.random.randint(0, high=df_max)

    D = {}
    D["t"] = t
    D["df"] = df
    D["k"] = k

    if mask_dim is not None:
        mask = np.random.choice(a=[False, True], size=mask_dim)
        D["mask"] = mask

    if dfs_max is not None:
        dfs = np.random.randint(1, dfs_max, size=t_dim)
        D["dfs"] = dfs

    if tri_dim is not None:
        tri = np.random.randint(1, np.size(t) - 1, size=tri_dim)
        D["tri"] = tri

    if resl_dim is not None:
        resl = np.random.random_sample(resl_dim)
        D["resl"] = resl

    if du is not None:
        D["du"] = du

    if c_dim is not None:
        c = np.random.random_sample(c_dim)
        D["c"] = c

    if ef_dim is not None:
        ef = np.random.random_sample(ef_dim)
        D["ef"] = ef

    if sd_dim is not None:
        sd = np.random.random_sample(sd_dim)
        D["sd"] = sd

    if X_dim is not None:
        first_col = np.ones(X_dim)
        seco_col = np.random.randint(10, 70, size=X_dim)
        X = np.concatenate((first_col, seco_col), axis=1)
        D["X"] = X

    if coef_dim is not None:
        coef = np.random.uniform(low=-2, high=2, size=coef_dim)
        D["coef"] = coef

    if SSE_dim is not None:
        SSE = np.random.uniform(low=0, high=3, size=SSE_dim)
        D["SSE"] = SSE

    with open(finname, "wb") as handle:
        pickle.dump(D, handle, protocol=4)

    return D


def get_fdr_output(D, foutname):
    """Runs fdr and returns all relevant output."""

    slm = SLM(FixedEffect(1), FixedEffect(1))
    for key in D.keys():
        setattr(slm, key, D[key])

    # run fdr
    Q = fdr(slm)

    Q_out = {}
    Q_out["Q"] = Q

    with open(foutname, "wb") as handle:
        pickle.dump(Q_out, handle, protocol=4)  #

    return


def generate_data_test_fdr():
    ### test_01 data in-out generation
    print("test_fdr: test_01 data is generated..")
    # random data shape matching a real-data set
    # ['t'] : np array, shape (1, 64984), float64
    # ['df'] : int
    # ['k'] : int
    t_dim = (1, 64984)
    df_max = 64984
    finname = datadir("xstatq_01_IN.pkl")
    D = generate_random_fdr_data(t_dim, df_max, finname, seed=444)
    foutname = datadir("xstatq_01_OUT.pkl")
    get_fdr_output(D, foutname)

    ### test_02 data in-out generation
    print("test_fdr: test_02 data is generated..")
    # random data
    # ['t'] : np array, shape (1, 9850), float64
    # ['df'] : int
    # ['k'] : int
    t_dim = (1, 9850)
    df_max = 1000
    finname = datadir("xstatq_02_IN.pkl")
    D = generate_random_fdr_data(t_dim, df_max, finname, seed=445)
    foutname = datadir("xstatq_02_OUT.pkl")
    get_fdr_output(D, foutname)

    ### test_03 data in-out generation
    print("test_fdr: test_03 data is generated..")
    # similar to test_02, shapes/values of slm['t'] and slm['df'] manipulated
    # ['t'] :  np array, shape (1, 2139), float64
    # ['df'] : int
    # ['k'] :  int
    t_dim = (1, 2139)
    df_max = 2000
    k = 3
    finname = datadir("xstatq_03_IN.pkl")
    D = generate_random_fdr_data(t_dim, df_max, finname, k=k, seed=446)
    foutname = datadir("xstatq_03_OUT.pkl")
    get_fdr_output(D, foutname)

    ### test_04 data in-out generation
    print("test_fdr: test_04 data is generated..")
    # similar to test_02 + optional input ['mask']
    # ['t'] : np array, shape (1, 2475), float64
    # ['df'] : int
    # ['k'] : int
    # ['mask'] : np array, shape (2475,), bool
    t_dim = (1, 2475)
    df_max = 1500
    finname = datadir("xstatq_04_IN.pkl")
    mask_dim = 2475
    D = generate_random_fdr_data(t_dim, df_max, finname, mask_dim=mask_dim, seed=447)
    foutname = datadir("xstatq_04_OUT.pkl")
    get_fdr_output(D, foutname)

    ### test_05 data in-out generation
    print("test_fdr: test_05 data is generated..")
    # similar to test_02 + optional input slm['dfs']
    # ['t'] : np array, shape (1, 1998), float64
    # ['df'] : int
    # ['k'] : int
    # ['dfs'] : np array, shape (1, 1998), int64
    t_dim = (1, 1998)
    df_max = 4000
    dfs_max = 1997
    finname = datadir("xstatq_05_IN.pkl")
    D = generate_random_fdr_data(t_dim, df_max, finname, dfs_max=dfs_max, seed=448)
    foutname = datadir("xstatq_05_OUT.pkl")
    get_fdr_output(D, foutname)

    ### test_06 data in-out generation
    print("test_fdr: test_06 data is generated..")
    # similar to test_02 + optional inputs slm['dfs'] and ['mask']
    # ['t'] : np array, shape (1, 3328), float64
    # ['df'] : np array, shape (1, 1), int64
    # ['k'] : int
    # ['dfs'] : np array, shape (1, 3328), int64
    # ['mask'] : np array, shape (3328,), bool
    t_dim = (1, 3328)
    df_max = 10000
    k = 2
    dfs_max = 3328
    mask_dim = 3328
    finname = datadir("xstatq_06_IN.pkl")
    D = generate_random_fdr_data(
        t_dim, df_max, finname, k=k, dfs_max=dfs_max, mask_dim=mask_dim, seed=449
    )
    foutname = datadir("xstatq_06_OUT.pkl")
    get_fdr_output(D, foutname)

    ### test_07 data in-out generation
    print("test_fdr: test_07 data is generated..")
    # similar to test_02 + optional inputs slm['dfs'], ['mask'] and ['tri']
    # ['t'] : np array, shape (1, 9512), float64
    # ['df'] : int
    # ['k'] : int
    # ['dfs'] : np array, shape (1, 9512), int64
    # ['mask'] : np array, shape (9512,), bool
    # ['tri'] : np array, shape (1724, 3), int64
    t_dim = (1, 9512)
    df_max = 5000
    dfs_max = 9511
    mask_dim = 9512
    tri_dim = (1724, 3)
    finname = datadir("xstatq_07_IN.pkl")
    D = generate_random_fdr_data(
        t_dim,
        df_max,
        finname,
        dfs_max=dfs_max,
        mask_dim=mask_dim,
        tri_dim=tri_dim,
        seed=450,
    )
    foutname = datadir("xstatq_07_OUT.pkl")
    get_fdr_output(D, foutname)

    ### test_08 data in-out generation
    print("test_fdr: test_08 data is generated..")
    # similar to test_02 + optional inputs slm['dfs'], slm['tri'] and slm['resl']
    # ['t'] : np array, shape (1, 1520), float64
    # ['df'] : int
    # ['k'] : int
    # ['dfs'] : np array, shape (1, 1520), int64
    # ['tri'] : np array, shape (4948, 3), int64
    # ['resl'] : np array, shape (1520, 1), float64
    t_dim = (1, 1520)
    df_max = 5000
    k = 5
    dfs_max = 9
    tri_dim = (4948, 3)
    resl_dim = (1520, 1)
    finname = datadir("xstatq_08_IN.pkl")
    D = generate_random_fdr_data(
        t_dim,
        df_max,
        finname,
        k=k,
        dfs_max=dfs_max,
        tri_dim=tri_dim,
        resl_dim=resl_dim,
        seed=451,
    )
    foutname = datadir("xstatq_08_OUT.pkl")
    get_fdr_output(D, foutname)

    ### test_09 data in-out generation
    print("test_fdr: test_09 data is generated..")
    # similar to test_08 + values/shapes of input params changed +
    # additional input slm['du'] (non-sense for _fdr)
    # ['t'] : np array, shape (1, 4397), float64
    # ['df'] : int
    # ['k'] : int
    # ['tri'] : np array, shape (2734, 3), int64
    # ['resl'] : np array, shape (8199, 1), float64
    # ['dfs'] : np array, shape (1, 4397), float64
    # ['du'] : int
    t_dim = (1, 14397)
    df_max = 1
    dfs_max = 2
    tri_dim = (2734, 3)
    resl_dim = (8199, 1)
    # du = 9
    finname = datadir("xstatq_09_IN.pkl")
    D = generate_random_fdr_data(
        t_dim,
        df_max,
        finname,
        dfs_max=dfs_max,
        tri_dim=tri_dim,
        resl_dim=resl_dim,  # du = du,
        seed=452,
    )
    foutname = datadir("xstatq_09_OUT.pkl")
    get_fdr_output(D, foutname)

    ### test_10 data in-out generation
    print("test_fdr: test_10 data is generated..")
    # similar to test_08 + + values/shapes of input params changed + additional
    # input slm['du'], slm['c'], slm['ef'], and slm['sd'] (non-sense for _fdr)
    # ['t'] : np array, shape (1, 20484), float64
    # ['df'] : int
    # ['k'] : int
    # ['tri'] : np array, shape (40960, 3), int32
    # ['resl'] : np array, shape (61440, 1), float64
    # ['c'] : np array, shape (1, 2), float64
    # ['ef'] : np array, shape (1, 20484), float64
    # ['sd'] : np array, shape (1, 20484), float64
    t_dim = (1, 20484)
    df_max = 10
    tri_dim = (40960, 3)
    resl_dim = (61440, 1)
    c_dim = (1, 2)
    ef_dim = (1, 20484)
    sd_dim = (1, 20484)
    finname = datadir("xstatq_10_IN.pkl")
    D = generate_random_fdr_data(
        t_dim,
        df_max,
        finname,
        tri_dim=tri_dim,
        resl_dim=resl_dim,
        c_dim=c_dim,
        ef_dim=ef_dim,
        sd_dim=sd_dim,
        seed=453,
    )
    foutname = datadir("xstatq_10_OUT.pkl")
    get_fdr_output(D, foutname)

    ### test_11 data in-out generation
    print("test_fdr: test_11 data is generated..")
    # similar to test_08 + additional input ['c'], ['ef'], ['sd'], ['X'],
    # and ['coef'], ['SSE'] (non-sense for _fdr)
    # ['t'] : np array, shape (1, 20484), float64
    # ['df'] : int
    # ['k'] : int
    # ['tri'] : np array, shape (40960, 3), int32
    # ['resl'] : np array, shape (61440, 1), float64
    # ['c'] : np array, shape (1, 2), float64
    # ['ef'] : np array, shape (1, 20484), float64
    # ['sd'] : np array, shape (1, 20484), float64
    # ['X'] : np array, shape (10, 2), float64
    # ['coef'] : np array, shape (2, 20484), float64
    # ['SSE'] : np array, shape (1, 20484), float64
    t_dim = (1, 20484)
    df_max = 10
    tri_dim = (40960, 3)
    resl_dim = (61440, 1)
    c_dim = (1, 2)
    ef_dim = (1, 20484)
    sd_dim = (1, 20484)
    X_dim = (10, 2)
    coef_dim = (2, 20484)
    SSE_dim = (1, 20484)
    finname = datadir("xstatq_11_IN.pkl")
    D = generate_random_fdr_data(
        t_dim,
        df_max,
        finname,
        tri_dim=tri_dim,
        resl_dim=resl_dim,
        c_dim=c_dim,
        ef_dim=ef_dim,
        sd_dim=sd_dim,
        X_dim=X_dim,
        coef_dim=coef_dim,
        SSE_dim=SSE_dim,
        seed=454,
    )
    foutname = datadir("xstatq_11_OUT.pkl")
    get_fdr_output(D, foutname)

    ### test_12 data in-out generation
    print("test_fdr: test_12 data is generated..")
    # similar to test_11 + optional input ['mask'] + ['df'] dtype changed
    # ['t'] : np array, shape (1, 20484), float64
    # ['df'] : uint8
    # ['k'] : int
    # ['tri'] : np array, shape (40960, 3), int32
    # ['resl'] : np array, shape (61440, 1), float64
    # ['c'] : np array, shape (1, 2), float64
    # ['ef'] : np array, shape (1, 20484), float64
    # ['sd'] : np array, shape (1, 20484), float64
    # ['X'] : np array, shape (10, 2), uint8
    # ['coef'] : np array, shape (2, 20484), float64
    # ['SSE'] : np array, shape (1, 20484), float64
    # ['mask'] : np array, shape (20484,), bool
    t_dim = (1, 20484)
    df_max = 10
    tri_dim = (40960, 3)
    resl_dim = (61440, 1)
    c_dim = (1, 2)
    ef_dim = (1, 20484)
    sd_dim = (1, 20484)
    X_dim = (10, 2)
    coef_dim = (2, 20484)
    SSE_dim = (1, 20484)
    mask_dim = 20484
    finname = datadir("xstatq_12_IN.pkl")
    D = generate_random_fdr_data(
        t_dim,
        df_max,
        finname,
        tri_dim=tri_dim,
        mask_dim=mask_dim,
        resl_dim=resl_dim,
        c_dim=c_dim,
        ef_dim=ef_dim,
        sd_dim=sd_dim,
        X_dim=X_dim,
        coef_dim=coef_dim,
        SSE_dim=SSE_dim,
        seed=455,
    )
    foutname = datadir("xstatq_12_OUT.pkl")
    get_fdr_output(D, foutname)

    ### test_13 data in-out generation
    print("test_fdr: test_13 data is generated..")
    # similar to test_10 + mask added
    # ['t'] : np array, shape (1, 20484), float64
    # ['df'] : int64
    # ['k'] : int
    # ['tri'] : np array, shape (40960, 3), int32
    # ['resl'] : np array, shape (61440, 1), float64
    # ['c'] : np array, shape (1, 9), float64
    # ['ef'] : np array, shape (1, 20484), float64
    # ['sd'] : np array, shape (1, 20484), float64
    # ['X'] : np array, shape (20, 9), uint16
    # ['coef'] : np array, shape (9, 20484), float64
    # ['SSE'] : np array, shape (1, 20484), float64
    t_dim = (1, 20484)
    df_max = 10
    tri_dim = (40960, 3)
    resl_dim = (61440, 1)
    c_dim = (1, 2)
    ef_dim = (1, 20484)
    sd_dim = (1, 20484)
    mask_dim = 20484
    finname = datadir("xstatq_13_IN.pkl")
    D = generate_random_fdr_data(
        t_dim,
        df_max,
        finname,
        tri_dim=tri_dim,
        resl_dim=resl_dim,
        c_dim=c_dim,
        mask_dim=mask_dim,
        ef_dim=ef_dim,
        sd_dim=sd_dim,
        seed=453,
    )
    foutname = datadir("xstatq_13_OUT.pkl")
    get_fdr_output(D, foutname)

    #### test 14, real data
    print("test_fdr: test_14 data is generated..")
    # thickness_n10 data, slm and t_test run prior to fdr
    realdataf = datadir("thickness_n10.pkl")
    ifile = open(realdataf, "br")
    DD = pickle.load(ifile)
    ifile.close()
    # run slm
    M = FixedEffect(DD["M"])
    slm = SLM(M, FixedEffect(1))
    slm.linear_model(DD["Y"])
    D = {}
    # run t-test
    t_test(slm)
    D["t"] = slm.t
    D["df"] = 10
    D["k"] = 1
    finname = datadir("xstatq_14_IN.pkl")
    with open(finname, "wb") as handle:
        pickle.dump(D, handle, protocol=4)
    foutname = datadir("xstatq_14_OUT.pkl")
    get_fdr_output(D, foutname)


if __name__ == "__main__":
    generate_data_test_fdr()



"""Data generation for mesh_smooth unit tests."""

import pickle

import numpy as np
from brainspace.mesh.mesh_elements import get_cells
from nilearn import datasets

from brainstat.context.utils import read_surface_gz
from brainstat.mesh.data import mesh_smooth
from brainstat.tests.testutil import datadir


def generate_smooth_out(rand_dict):
    """Uses rand_dict to run mesh_smooth and returns the smoothed data."""
    # below are going to be input params for mesh_smooth
    Y = rand_dict["Y"]
    FWHM = rand_dict["FWHM"]
    surf = {}
    if "tri" in rand_dict.keys():
        surf["tri"] = rand_dict["tri"]
    if "lat" in rand_dict.keys():
        surf["lat"] = rand_dict["lat"]

    # run mesh_smooth and return the smoothed data
    O = {}
    O["Python_Y"] = mesh_smooth(Y, surf, FWHM)

    return O


def params2files(rand_dict, O, test_num):
    """Converts params to input/output files"""
    # filenames for the input and outpur dict's
    basename = "xstatsmo"
    fin_name = datadir(basename + "_" + f"{test_num:02d}" + "_IN.pkl")
    fout_name = datadir(basename + "_" + f"{test_num:02d}" + "_OUT.pkl")

    # save input and output data in pickle format with filenames above
    with open(fin_name, "wb") as f:
        pickle.dump(rand_dict, f, protocol=4)
    with open(fout_name, "wb") as g:
        pickle.dump(O, g, protocol=4)
    return


# Single test with real data.
# There's no forks in this function other than the tri/lat which is already
# tested through the mesh_edges tests. We have to use real data as random data
# is likely to result in nan arrays due to faulty fake meshes.
def generate_tests():
    pial_fs5 = datasets.fetch_surf_fsaverage()["pial_left"]
    pial_surf = read_surface_gz(pial_fs5)
    tri = np.array(get_cells(pial_surf)) + 1

    np.random.seed(0)
    data = {"tri": tri, "Y": np.random.uniform(-1, 1, (1, 10242)), "FWHM": 3}
    O = generate_smooth_out(data)
    params2files(data, O, 1)


if __name__ == "__main__":
    generate_tests()



"""Genetic decoding using abagen."""
import tempfile
from pathlib import Path
from typing import Optional, Sequence, Union

import nibabel as nib
import numpy as np
import pandas as pd
from abagen import check_atlas, get_expression_data
from brainspace.mesh.mesh_io import read_surface, write_surface
from sklearn.model_selection import ParameterGrid

from brainstat._utils import data_directories, logger
from brainstat.datasets.base import (
    _fetch_template_surface_files,
    _valid_parcellations,
    fetch_parcellation,
)


def surface_genetic_expression(
    labels: Union[Sequence[str], np.ndarray],
    surfaces: Optional[Union[str, Sequence[str]]] = None,
    space: Optional[str] = None,
    *,
    atlas_info: str = None,
    ibf_threshold: float = 0.5,
    probe_selection: str = "diff_stability",
    donor_probes: str = "aggregate",
    lr_mirror: Optional[bool] = None,
    missing: Optional[str] = None,
    tolerance: float = 2,
    sample_norm: str = "srs",
    gene_norm: str = "srs",
    norm_matched: bool = True,
    norm_structures: bool = False,
    region_agg: str = "donors",
    agg_metric: str = "mean",
    corrected_mni: bool = True,
    reannotated: bool = True,
    return_counts: bool = False,
    return_donors: bool = False,
    return_report: bool = False,
    donors: str = "all",
    data_dir: Optional[str] = None,
    verbose: float = 0,
    n_proc: int = 1
) -> pd.DataFrame:
    """Computes genetic expression of surface parcels.

    Parameters
    ----------
    labels : list-of-str or numpy.ndarray
        List of paths to label files for the parcellation, or numpy array
        containing the pre-loaded labels
    surfaces : list-of-image, optional
        List of paths to surface files or preloaded surfaces. If not specified
        assumes that `labels` are on the `fsaverage5` surface. Default: None
    space : {'fsaverage', 'fslr'}
        What template space `surfaces` are aligned to. If not specified assumes
        that `labels` are on the `fsaverage5` surface. Default: None

    For details of the remaining parameters please consult the
    abagen.get_expression_data() documentation. All its parameters bar "atlas"
    are valid input parameters.

    Returns
    -------
    pandas.DataFrame
        Dataframe containing the expression of each gene within each region.

    Examples
    --------
    >>> from brainstat.context.genetics import surface_genetic_expression
    >>> from nilearn import datasets
    >>> import numpy as np

    >>> destrieux = datasets.fetch_atlas_surf_destrieux()
    >>> labels = np.hstack((destrieux['map_left'], destrieux['map_right']))
    >>> fsaverage = datasets.fetch_surf_fsaverage()
    >>> surfaces = (fsaverage['pial_left'], fsaverage['pial_right'])
    >>> expression = surface_genetic_expression(labels, surfaces,
    ...                                         space='fsaverage')
    """

    # Deal with the input parameters.
    if isinstance(surfaces, str):
        surfaces = [surfaces]
    elif surfaces is None:
        surfaces = []

    surfaces_gii = []
    for surface in surfaces:
        if not isinstance(surface, str) and not isinstance(surface, Path):
            # Rather roundabout deletion of the temporary file for Windows compatibility.
            try:
                with tempfile.NamedTemporaryFile(suffix=".gii", delete=False) as f:
                    name = f.name
                    write_surface(surface, name, otype="gii")
                surfaces_gii.append(nib.load(name))
            finally:
                Path(name).unlink()
        else:
            surfaces_gii.append(nib.load(surface))

    # Use abagen to grab expression data.
    logger.info(
        "If you use BrainStat's genetics functionality, please cite abagen (https://abagen.readthedocs.io/en/stable/citing.html)."
    )
    atlas = check_atlas(labels, geometry=surfaces_gii, space=space)
    expression = get_expression_data(
        atlas,
        atlas_info=atlas_info,
        ibf_threshold=ibf_threshold,
        probe_selection=probe_selection,
        donor_probes=donor_probes,
        lr_mirror=lr_mirror,
        missing=missing,
        tolerance=tolerance,
        sample_norm=sample_norm,
        gene_norm=gene_norm,
        norm_matched=norm_matched,
        norm_structures=norm_structures,
        region_agg=region_agg,
        agg_metric=agg_metric,
        corrected_mni=corrected_mni,
        reannotated=reannotated,
        return_counts=return_counts,
        return_donors=return_donors,
        return_report=return_report,
        donors=donors,
        data_dir=data_dir,
        verbose=verbose,
        n_proc=n_proc,
    )

    return expression


def __create_precomputed(
    data_dir: Optional[Union[str, Path]] = None,
    output_dir: Optional[Union[str, Path]] = None,
) -> None:
    """Creates precomputed matrices of genetic expression for standard atlases. These are
    used for the MATLAB implementation."""

    if output_dir is None:
        output_dir = data_dir = (
            Path(data_dir) if data_dir else data_directories["BRAINSTAT_DATA_DIR"]
        )

    # Get all parcellations in a format compatible with ParameterGrid.
    parcellations = _valid_parcellations()
    for key in parcellations:
        parcellations[key].update({"atlas": [key]})
        if key == "schaefer":
            parcellations[key].update({"seven_networks": (True, False)})
        else:
            parcellations[key].update({"seven_networks": (True,)})

        # We only really need to compute this for one template surface.
        # We'll use fsaverage for all.
        parcellations[key].update({"surfaces": ["fsaverage"]})

    param_grid = ParameterGrid(list(parcellations.values()))

    # Compute expression for all parcellations.
    for params in param_grid:
        surface_files = _fetch_template_surface_files(
            params["surfaces"], data_dir=data_dir
        )
        space = "fslr" if params["surfaces"] == "fslr32k" else "fsaverage"

        labels = fetch_parcellation(
            params["surfaces"],
            params["atlas"],
            params["n_regions"],
            seven_networks=params["seven_networks"],
            data_dir=data_dir,
        )

        surf_lh = tempfile.NamedTemporaryFile(suffix=".surf.gii")
        surf_rh = tempfile.NamedTemporaryFile(suffix=".surf.gii")
        for i, surf in enumerate((surf_lh, surf_rh)):
            __freesurfer_to_surfgii(surface_files[i], surf.name)
        surface_paths = [surf_lh.name, surf_rh.name]

        expression = surface_genetic_expression(labels, surface_paths, space=space)  # type: ignore

        if params["atlas"] == "schaefer":
            network_tag = "7Networks" if params["seven_networks"] else "17Networks"
        else:
            network_tag = ""

        filename_components = filter(
            None,
            (
                "expression",
                params["atlas"],
                str(params["n_regions"]),
                network_tag,
            ),
        )
        filename = "_".join(filename_components) + ".csv.gz"
        expression.to_csv(Path(output_dir, filename))  # type: ignore


def __freesurfer_to_surfgii(freesurfer_file: str, gifti_file: str) -> None:
    surf = read_surface(freesurfer_file, itype="fs")
    write_surface(surf, gifti_file, otype="gii")



import shutil
import warnings
from pathlib import Path
from typing import Optional, Sequence, Tuple, Union
from urllib.error import HTTPError
from urllib.request import urlopen

import numpy as np
import pandas as pd
from sklearn.utils import Bunch
from tqdm import tqdm

from brainstat._utils import data_directories, read_data_fetcher_json


def fetch_abide_data(
    data_dir: Optional[Union[str, Path]] = None,
    sites: Sequence[str] = None,
    keep_control: bool = True,
    keep_patient: bool = True,
    overwrite: bool = False,
) -> Tuple[Bunch, Bunch]:

    data_dir = Path(data_dir) if data_dir else data_directories["ABIDE_DATA_DIR"]
    data_dir.mkdir(exist_ok=True, parents=True)
    summary_spreadsheet = data_dir / "summary_spreadsheet.csv"

    if not summary_spreadsheet.is_file():
        summary_url = read_data_fetcher_json()["abide_tutorial"]["summary_spreadsheet"]
        _download_file(summary_spreadsheet, summary_url["url"])

    df = pd.read_csv(summary_spreadsheet)
    _select_subjects(df, sites, keep_patient, keep_control)

    # Download subject thickeness data
    def _thickness_url(derivative, identifier):
        return f"https://s3.amazonaws.com/fcp-indi/data/Projects/ABIDE_Initiative/Outputs/civet/thickness_{derivative}/{identifier}_{derivative}.txt"

    thickness_data = np.zeros((df.shape[0], 81924))
    remove_rows = []
    progress_bar = tqdm(df.itertuples())
    for i, row in enumerate(progress_bar):
        progress_bar.set_description(
            f"Fetching thickness data for subject {i+1} out of {df.shape[0]}"
        )
        for j, hemi in enumerate(["left", "right"]):
            filename = data_dir / f"{row.SUB_ID}_{hemi}_thickness.txt"
            if not filename.is_file() or overwrite:
                thickness_url = _thickness_url(
                    f"native_rms_rsl_tlink_30mm_{hemi}", row.FILE_ID
                )
                try:
                    _download_file(filename, thickness_url)
                except HTTPError:
                    warnings.warn(f"Could not download file for {row.SUB_ID}.")
                    remove_rows.append(i)
                    continue

            thickness_data[i, j * 40962 : (j + 1) * 40962] = np.loadtxt(filename)

    if remove_rows:
        thickness_data = np.delete(thickness_data, remove_rows, axis=0)
        df.drop(np.unique(remove_rows), inplace=True)
        df.reset_index(inplace=True)

    return thickness_data, df


def _download_file(filename: Path, url: str) -> None:
    if not filename.is_file():
        with urlopen(url) as u, open(filename, "wb") as f:
            shutil.copyfileobj(u, f)


def _select_subjects(
    df: pd.DataFrame,
    sites: Optional[Sequence[str]],
    keep_patient: bool,
    keep_control: bool,
) -> None:
    df.drop(df[df.FILE_ID == "no_filename"].index, inplace=True)
    if not keep_patient:
        df.drop(df[df.DX_GROUP == 1].index, inplace=True)

    if not keep_control:
        df.drop(df[df.DX_GROUP == 2].index, inplace=True)

    if sites is not None:
        df.drop(df[~df.SITE_ID.isin(sites)].index, inplace=True)

    df.reset_index(inplace=True)



"""Unit tests of SLM."""
import pickle

import numpy as np
import pandas as pd
import pytest

from brainstat.stats.SLM import SLM, _onetailed_to_twotailed
from brainstat.stats.terms import FixedEffect, MixedEffect
from brainstat.tests.testutil import datadir


def recursive_comparison(X1, X2):
    """Recursively compares lists/dictionaries."""

    if type(X1) != type(X2):
        raise ValueError("Both inputs must be of the same type.")

    if isinstance(X1, dict):
        if len(X1.keys()) != len(X2.keys()):
            raise ValueError("Different number of keys in each dictionary.")
        iterator = zip(X1.values(), X2.values())
    elif isinstance(X1, list):
        if len(X1) != len(X2):
            raise ValueError("Different number of elements in each list.")
        iterator = zip(X1, X2)
    elif isinstance(X1, pd.DataFrame):
        iterator = zip(X1.values.tolist(), X2.values.tolist())
    else:
        # Assume not iterable.
        iterator = zip([X1], [X2])

    output = True
    for x, y in iterator:
        try:
            if x is None and y is None:
                output = True
            elif (
                isinstance(x, list)
                or isinstance(x, dict)
                or isinstance(x, pd.DataFrame)
            ):
                output = recursive_comparison(x, y)
            else:
                output = np.allclose(x, y)
            if not output:
                return output
        except:
            import pdb

            pdb.set_trace()
    return output


def dummy_test(infile, expfile):

    # load input test data
    ifile = open(infile, "br")
    idic = pickle.load(ifile)
    ifile.close()

    slm = SLM(FixedEffect(1), FixedEffect(1))
    # Data are saved a little differently from the actual input due to compatibility with MATLAB.
    # Data wrangle a bit to bring it back into the Python input format.
    for key in idic.keys():
        if key == "Y":
            # Y is input for slm.fit(), not a property.
            continue
        if key == "model":
            # Model is saved as a matrix rather than a Fixed/MixedEffect
            if idic[key].shape[1] == 1:
                idic[key] = FixedEffect(1) + FixedEffect(idic[key])
            else:
                idic[key] = (
                    FixedEffect(1)
                    + FixedEffect(idic[key][:, 0])
                    + MixedEffect(idic[key][:, 1])
                    + MixedEffect(1)
                )
        setattr(slm, key, idic[key])
        if key == "surf" and slm.surf is not None:
            slm.surf["tri"] += 1

    slm.fit(idic["Y"])

    # load expected outout data
    efile = open(expfile, "br")
    out = pickle.load(efile)
    efile.close()

    # Format of self.P changed since files were created -- alter out to match some changes.
    # Combine the list outputs, sort with pandas, and return to list.
    if "P" in out:
        out["P"]["pval"]["C"] = _onetailed_to_twotailed(
            out["P"]["pval"]["C"][0], out["P"]["pval"]["C"][1]
        )

        for key1 in ["peak", "clus"]:
            P_tmp = []
            none_squeeze = lambda x: np.squeeze(x) if x is not None else None
            for i in range(len(out["P"][key1]["P"])):
                tail_dict = {
                    key: none_squeeze(value[i]) for key, value in out["P"][key1].items()
                }
                if tail_dict["P"] is not None:
                    if tail_dict["P"].size == 1:
                        P_tmp.append(pd.DataFrame.from_dict([tail_dict]))
                    else:
                        P_tmp.append(pd.DataFrame.from_dict(tail_dict))
                        P_tmp[i].sort_values(by="P", ascending=True)
                else:
                    P_tmp.append(pd.DataFrame(columns=tail_dict.keys()))
            out["P"][key1] = P_tmp

    testout = []

    skip_keys = ["model", "correction", "_tri", "surf"]

    for key in out.keys():
        if key in skip_keys:
            continue
        if key == "P":
            testout.append(recursive_comparison(out[key], getattr(slm, key)))
        elif out[key] is not None:
            comp = np.allclose(out[key], getattr(slm, key), rtol=1e-05, equal_nan=True)
            testout.append(comp)

    assert all(flag == True for (flag) in testout)


expected_number_of_tests = 22
parametrize = pytest.mark.parametrize


@parametrize("test_number", range(1, expected_number_of_tests + 1))
def test_run_all(test_number):
    infile = datadir("slm_" + f"{test_number:02d}" + "_IN.pkl")
    expfile = datadir("slm_" + f"{test_number:02d}" + "_OUT.pkl")
    dummy_test(infile, expfile)



"""Interpolations on a mesh."""
import gzip
import os
import shutil
import tempfile
from pathlib import Path
from typing import List, Optional, Sequence, Union

import nibabel as nib
import numpy as np
import trimesh
from brainspace.mesh.mesh_elements import get_cells, get_points
from brainspace.mesh.mesh_io import read_surface
from brainspace.vtk_interface.wrappers.data_object import BSPolyData
from nilearn.datasets import load_mni152_brain_mask
from scipy.interpolate.ndgriddata import LinearNDInterpolator, NearestNDInterpolator
from scipy.spatial import cKDTree

from brainstat._utils import data_directories, logger
from brainstat.datasets import fetch_template_surface


def surface_to_volume(
    pial_mesh: Union[str, BSPolyData],
    wm_mesh: Union[str, BSPolyData],
    labels: Union[str, np.ndarray],
    volume_template: Union[str, nib.nifti1.Nifti1Image],
    volume_save: str,
    interpolation: str = "nearest",
) -> None:
    """Projects surface labels to the cortical ribbon.

    Parameters
    ----------
    pial_mesh : str, BSPolyData
        Filename of a pial mesh or a BSPolyData object of the same.
    wm_mesh : str, BSPolyData
        Filename of a pial mesh or a BSPolyData object of the same.
    labels : str, numpy.ndarray
        Filename of a .label.gii or .shape.gii file, or a numpy array
        containing the labels.
    volume_template : str, nibabel.nifti1.Nifti1Image
        Filename of a nifti image in the same space as the mesh files or a
        NIfTI image loaded with nibabel.
    volume_save : str
        Filename to which the label image will be saved.
    interpolation : str
        Either 'nearest' for nearest neighbor interpolation, or 'linear'
        for trilinear interpolation, defaults to 'nearest'.
    """

    if not isinstance(pial_mesh, BSPolyData):
        pial_mesh = read_surface(pial_mesh)
    if not isinstance(wm_mesh, BSPolyData):
        wm_mesh = read_surface(wm_mesh)
    if not isinstance(volume_template, nib.nifti1.Nifti1Image):
        volume_template = nib.load(volume_template)

    logger.debug("Computing voxels inside the cortical ribbon.")
    ribbon_points = cortical_ribbon(pial_mesh, wm_mesh, volume_template)

    logger.debug("Computing labels for cortical ribbon voxels.")
    ribbon_labels = ribbon_interpolation(
        pial_mesh,
        wm_mesh,
        labels,
        volume_template,
        ribbon_points,
        interpolation=interpolation,
    )

    logger.debug("Constructing new nifti image.")
    new_data = np.zeros(volume_template.shape)
    ribbon_points = np.rint(
        ribbon_points, np.ones(ribbon_points.shape, dtype=int), casting="unsafe"
    )
    for i in range(ribbon_labels.shape[0]):
        new_data[
            ribbon_points[i, 0], ribbon_points[i, 1], ribbon_points[i, 2]
        ] = ribbon_labels[i]

    new_nii = nib.Nifti1Image(new_data, volume_template.affine)
    nib.save(new_nii, volume_save)


valid_surfaces = Union[
    str,
    BSPolyData,
    Sequence[Union[str, BSPolyData]],
]


def _input_to_list(x: valid_surfaces) -> List[Union[str, BSPolyData]]:
    if isinstance(x, str):
        return [x]
    else:
        return list(x)


def multi_surface_to_volume(
    pial: valid_surfaces,
    white: valid_surfaces,
    volume_template: Union[str, nib.nifti1.Nifti1Image],
    output_file: str,
    labels: Union[str, np.ndarray, Sequence[Union[np.ndarray, str]]],
    interpolation: str = "nearest",
) -> None:
    """Interpolates multiple surfaces to the volume.

    Parameters
    ----------
    pial : str, BSPolyData, list, tuple
        Path of a pial surface file, BSPolyData of a pial surface or a list
        containing multiple of the aforementioned.
    white : str, BSPolyData, list, tuple
        Path of a white matter surface file, BSPolyData of a pial surface or a
        list containing multiple of the aforementioned.
    volume_template : str, nibabel.nifti1.Nifti1Image
        Path to a nifti file to use as a template for the surface to volume
        procedure, or a loaded NIfTI image.
    output_file: str
        Path to the output file, must end in .nii or .nii.gz.
    labels : str, numpy.ndarray, list, tuple
        Path to a label file for the surfaces, numpy array containing the
        labels, or a list containing multiple of the aforementioned.
    interpolation : str
        Either 'nearest' for nearest neighbor interpolation, or 'linear'
        for trilinear interpolation, defaults to 'nearest'.

    Notes
    -----
    An equal number of pial/white surfaces and labels must be provided. If
    parcellations overlap across surfaces, then the labels are kept for the
    first provided surface.
    """

    # Deal with variety of ways to provide input.
    if type(pial) is not type(white):
        raise ValueError("Pial and white must be of the same type.")

    pial_list = _input_to_list(pial)
    white_list = _input_to_list(white)
    labels_list = _input_to_list(labels)

    if len(pial_list) is not len(white):
        raise ValueError("The same number of pial and white surfces must be provided.")

    for i in range(len(pial_list)):
        if not isinstance(pial_list[i], BSPolyData):
            pial_list[i] = read_surface_gz(pial_list[i])

        if not isinstance(white_list[i], BSPolyData):
            white_list[i] = read_surface_gz(white_list[i])

    if not isinstance(volume_template, nib.nifti1.Nifti1Image):
        volume_template = nib.load(volume_template)

    for i in range(len(labels_list)):
        if isinstance(labels_list[i], np.bool_):
            labels_list[i] = np.array(labels_list[i])
        elif not isinstance(labels[i], np.ndarray):
            labels_list[i] = load_mesh_labels(labels_list[i])

    # Surface data to volume.
    T = []
    for i in range(len(pial)):
        T.append(tempfile.NamedTemporaryFile(suffix=".nii.gz"))
        surface_to_volume(
            pial_list[i],
            white_list[i],
            labels[i],
            volume_template,
            T[i].name,
            interpolation=interpolation,
        )

    if len(T) > 1:
        T_names = [x.name for x in T]
        combine_parcellations(T_names, output_file)
    else:
        shutil.copy(T[0].name, output_file)


def cortical_ribbon(
    pial_mesh: BSPolyData,
    wm_mesh: BSPolyData,
    nii: nib.nifti1.Nifti1Image,
    mesh_distance: float = 6,
) -> np.ndarray:
    """Finds voxels inside of the cortical ribbon.

    Parameters
    ----------
    pial_mesh : BSPolyData
        Pial mesh.
    wm_mesh : BSPolyData
        White matter mesh.
    nii : Nibabel nifti
        Nifti image containing the space in which to output the ribbon.
    mesh_distance : float, optional
        Maximum distance from the cortical mesh at which the ribbon may occur.
        Used to reduce the search space, by default 6.

    Returns
    -------
    numpy.ndarray
        Matrix coordinates of voxels inside the cortical ribbon.
    """

    try:
        import pyembree
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "The package pyembree is required for this function. "
            + "You can install it with the conda package manager: "
            + "`conda install -c conda-forge pyembree`."
        )

    # Get world coordinates.
    x, y, z, _ = np.meshgrid(
        range(nii.shape[0]), range(nii.shape[1]), range(nii.shape[2]), 0
    )

    points = np.reshape(np.concatenate((x, y, z), axis=3), (-1, 3), order="F")
    world_coord = nib.affines.apply_affine(nii.affine, points)

    logger.debug("Discarding points that exceed the minima/maxima of the pial mesh.")
    # Discard points that exceed any of the maxima/minima
    pial_points = np.array(get_points(pial_mesh))
    discard = np.any(
        # If points exceed maximum coordinates
        (world_coord > np.amax(pial_points, axis=0)) |
        # If points are lower than minimum coordinates
        (world_coord < np.amin(pial_points, axis=0)),
        axis=1,
    )
    world_coord = world_coord[np.logical_not(discard), :]

    # Discard points that are more than mesh_distance from the pial and wm mesh.
    logger.debug("Discarding points that are too far from the meshes.")
    tree = cKDTree(pial_points)
    mindist_pial, _ = tree.query(world_coord)

    wm_points = np.array(get_points(wm_mesh))
    tree = cKDTree(wm_points)
    mindist_wm, _ = tree.query(world_coord)

    world_coord = world_coord[
        (mindist_pial < mesh_distance) & (mindist_wm < mesh_distance), :
    ]

    # Check which points are inside pial but not inside WM (i.e. ribbon)
    logger.debug("Retaining only points that are inside the pial but not the WM mesh.")
    pial_trimesh = trimesh.ray.ray_pyembree.RayMeshIntersector(
        trimesh.Trimesh(
            vertices=np.array(get_points(pial_mesh)),
            faces=np.array(get_cells(pial_mesh)),
        )
    )
    wm_trimesh = trimesh.ray.ray_pyembree.RayMeshIntersector(
        trimesh.Trimesh(
            vertices=np.array(get_points(wm_mesh)), faces=np.array(get_cells(wm_mesh))
        )
    )

    inside_wm = wm_trimesh.contains_points(world_coord)
    inside_pial = pial_trimesh.contains_points(world_coord)
    inside_ribbon = world_coord[inside_pial & ~inside_wm, :]
    ribbon_points = nib.affines.apply_affine(np.linalg.inv(nii.affine), inside_ribbon)
    return ribbon_points


def ribbon_interpolation(
    pial_mesh: BSPolyData,
    wm_mesh: BSPolyData,
    labels: Union[str, np.ndarray],
    nii: nib.nifti1.Nifti1Image,
    points: np.ndarray,
    interpolation: str = "nearest",
) -> np.ndarray:
    """Performs label interpolation in the cortical ribbon.

    Parameters
    ----------
    pial_mesh : BSPolyData
        Pial mesh.
    wm_mesh : BSPolydata
        White matter mesh.
    labels : str, numpy.ndarray
        Filename of a .label.gii or .shape.gii file, or a numpy array
        containing the labels.
    nii : Nibabel nifti
        Reference nifti image.
    points : numpy.array
        Numpy array containing the coordinates of the ribbon.
    interpolation : str, optional
        Interpolation method. Can be either 'nearest' or 'linear'.

    Returns
    -------
    numpy.ndarray
        Interpolated value for each input point.

    Notes
    -----
    Strictly, this function will work outside the cortical ribbon too and assign
    any point to its label on the nearest mesh. An adventurous user could use
    this for nearest neighbour surface to volume anywhere in the brain, although
    such usage is not offically supported.
    """

    if not isinstance(labels, np.ndarray):
        labels = nib.gifti.giftiio.read(labels).agg_data()

    mesh_coord = np.concatenate((get_points(pial_mesh), get_points(wm_mesh)), axis=0)

    # Repeat labels as we concatenate the pial/white meshes.
    labels = np.concatenate((labels, labels))

    ribbon_coord = nib.affines.apply_affine(nii.affine, points)

    if interpolation == "nearest":
        interp = NearestNDInterpolator(mesh_coord, labels)
    elif interpolation == "linear":
        interp = LinearNDInterpolator(mesh_coord, labels)
    else:
        ValueError("Unknown interpolation type.")

    return interp(ribbon_coord)


def __create_precomputed(data_dir: Optional[Union[str, Path]] = None) -> None:
    """Create nearest neighbor interpolation niftis for MATLAB."""
    data_dir = Path(data_dir) if data_dir else data_directories["BRAINSTAT_DATA_DIR"]
    mni152 = load_mni152_brain_mask()
    for template in ("fsaverage5", "fsaverage", "civet41k", "civet164k"):
        output_file = data_dir / f"nn_interp_{template}.nii.gz"
        if output_file.exists():
            continue
        top_surf = "pial" if template[:9] == "fsaverage" else "mid"
        pial = fetch_template_surface(template, layer=top_surf, join=False)
        white = fetch_template_surface(template, layer="white", join=False)
        labels = (
            np.arange(1, get_points(pial[0]).shape[0] + 1),
            np.arange(
                get_points(pial[0]).shape[0] + 1, get_points(pial[0]).shape[0] * 2 + 1
            ),
        )
        multi_surface_to_volume(
            pial=pial,
            white=white,
            volume_template=mni152,
            labels=labels,
            output_file=str(output_file),
            interpolation="nearest",
        )

    if not (data_dir / "nn_interp_hcp.nii.gz").exists():
        import hcp_utils as hcp
        from brainspace.mesh.mesh_creation import build_polydata

        pial_fslr32k = (build_polydata(hcp.mesh.pial[0], hcp.mesh.pial[1]),)
        white_fslr32k = (build_polydata(hcp.mesh.white[0], hcp.mesh.white[1]),)
        labels_fslr32k = (np.arange(1, get_points(pial_fslr32k[0]).shape[0] + 1),)
        multi_surface_to_volume(
            pial=pial_fslr32k,
            white=white_fslr32k,
            volume_template=mni152,
            labels=labels_fslr32k,
            output_file=str(data_dir / "nn_interp_fslr32k.nii.gz"),
            interpolation="nearest",
        )


def combine_parcellations(files: List[str], output_file: str) -> None:
    """Combines multiple nifti files into one.

    Parameters
    ----------
    files : list
        List of strings containing the paths to nifti files.
    output_file : str
        Path to the output file.

    Notes
    -----
    This function assumes that 0's are missing data. When multiple files have
    non-zero values in the same voxel, then the data from the first provided
    file is kept.
    """
    for i in range(len(files)):
        nii = nib.load(files[i])
        if i == 0:
            img = nii.get_fdata()
            affine = nii.affine
            header = nii.header
        else:
            img[img == 0] = nii.get_fdata()[img == 0]
    new_nii = nib.Nifti1Image(img, affine, header)
    nib.save(new_nii, output_file)


def load_mesh_labels(label_file: str, as_int: bool = True) -> np.ndarray:
    """Loads a .label.gii or .csv file.

    Parameters
    ----------
    label_file : str
        Path to the label file.
    as_int : bool
        Determines whether to enforce integer format on the labels, defaults to True.

    Returns
    -------
    numpy.ndarray
        Labels in the file.
    """

    if label_file.endswith(".gii"):
        labels = nib.gifti.giftiio.read(label_file).agg_data()
    elif label_file.endswith(".csv"):
        labels = np.loadtxt(label_file)
    else:
        raise ValueError("Unrecognized label file type.")

    if as_int:
        labels = np.round(labels).astype(int)
    return labels


def read_surface_gz(filename: str) -> BSPolyData:
    """Extension of brainspace's read_surface to include .gz files.

    Parameters
    ----------
    filename : str
        Filename of file to open.

    Returns
    -------
    BSPolyData
        Surface mesh.
    """
    if filename.endswith(".gz"):
        extension = os.path.splitext(filename[:-3])[-1]
        with tempfile.NamedTemporaryFile(suffix=extension) as f_tmp:
            with gzip.open(filename, "rb") as f_gz:
                shutil.copyfileobj(f_gz, f_tmp)
            return read_surface(f_tmp.name)
    else:
        return read_surface(filename)



"""Utilities for BrainStat developers."""
import json
import logging
import logging.config
import shutil
import urllib.request
import warnings
from pathlib import Path
from typing import Callable

import brainstat

json_file = Path(brainstat.__file__).parent / "data_urls.json"

logging.config.fileConfig(Path(brainstat.__file__).parent / "logging.conf")
logger = logging.getLogger("brainstat")


BRAINSTAT_DATA_DIR = Path.home() / "brainstat_data"
data_directories = {
    "BRAINSTAT_DATA_DIR": BRAINSTAT_DATA_DIR,
    "ABIDE_DATA_DIR": BRAINSTAT_DATA_DIR / "abide_data",
    "BIGBRAIN_DATA_DIR": BRAINSTAT_DATA_DIR / "bigbrain_data",
    "GRADIENT_DATA_DIR": BRAINSTAT_DATA_DIR / "gradient_data",
    "NEUROSYNTH_DATA_DIR": BRAINSTAT_DATA_DIR / "neurosynth_data",
    "PARCELLATION_DATA_DIR": BRAINSTAT_DATA_DIR / "parcellation_data",
    "SURFACE_DATA_DIR": BRAINSTAT_DATA_DIR / "surface_data",
}


def generate_data_fetcher_json() -> None:
    """Stores the URLS of all external data in a .json file."""
    data = {
        "bigbrain_profiles": {
            "fsaverage": {
                "url": "https://box.bic.mni.mcgill.ca/s/znBp7Emls0mMW1a/download",
            },
            "fsaverage5": {
                "url": "https://box.bic.mni.mcgill.ca/s/N8zstvuRb4sNcSe/download",
            },
            "fslr32k": {
                "url": "https://box.bic.mni.mcgill.ca/s/6zKHcg9xXu5inPR/download",
            },
        },
        "gradients": {
            "margulies2016": {
                "url": "https://box.bic.mni.mcgill.ca/s/LWFaQlOxUWmRlc0/download",
            }
        },
        "neurosynth_precomputed": {
            "url": "https://box.bic.mni.mcgill.ca/s/GvislmLffbCIZoI/download",
            "n_files": 3228,
        },
        "parcellations": {
            "glasser": {
                "fsaverage": {
                    "url": (
                        "https://box.bic.mni.mcgill.ca/s/y2NMHXr47WOCtpp/download",
                        "https://box.bic.mni.mcgill.ca/s/Y0Fmd2tIF69Mqpt/download",
                    ),
                },
                "fsaverage5": {
                    "url": (
                        "https://box.bic.mni.mcgill.ca/s/Kg4VdWRt4NHvr3B/download",
                        "https://box.bic.mni.mcgill.ca/s/9sEXgVKi3VJ9pXV/download",
                    ),
                },
                "fslr32k": {
                    "url": (
                        "https://box.bic.mni.mcgill.ca/s/y2NMHXr47WOCtpp/download",
                        "https://box.bic.mni.mcgill.ca/s/Y0Fmd2tIF69Mqpt/download",
                    ),
                },
            },
            "yeo": {"url": "https://box.bic.mni.mcgill.ca/s/vcSXEk1wx0jN86N/download"},
        },
        "masks": {
            "civet41k": {
                "url": "https://box.bic.mni.mcgill.ca/s/9kzBetBCZkkqN6w/download"
            },
            "civet164k": {
                "url": "https://box.bic.mni.mcgill.ca/s/rei5HtTDvexlEPA/download"
            },
        },
        "abide_tutorial": {
            "summary_spreadsheet": {
                "url": "https://s3.amazonaws.com/fcp-indi/data/Projects/ABIDE_Initiative/Phenotypic_V1_0b_preprocessed1.csv"
            },
        },
    }
    with open(json_file, "w") as f:
        json.dump(data, f, indent=4, sort_keys=True)


def read_data_fetcher_json() -> dict:
    """Reads the URLS of all external data from a .json file."""
    with open(json_file, "r") as f:
        return json.load(f)


def deprecated(message: str) -> Callable:
    """Decorator for deprecated functions.

    Parameters
    ----------
    message : str
        Message to return to the user.
    """

    def deprecated_decorator(func):
        def deprecated_func(*args, **kwargs):
            warnings.warn(
                "{} is a deprecated function and will be removed in a future version. {}".format(
                    func.__name__, message
                ),
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)
            return func(*args, **kwargs)

        return deprecated_func

    return deprecated_decorator


def _download_file(url: str, output_file: Path, overwrite: bool = False) -> None:
    """Downloads a file.

    Parameters
    ----------
    url : str
        URL of the download.
    output_file : pathlib.Path
        Path object of the output file.
    overwrite : bool
        If true, overwrite existing files, defaults to False.
    """

    if output_file.exists() and not overwrite:
        return

    logger.info("Downloading " + str(output_file) + " from " + url + ".")
    with urllib.request.urlopen(url) as response, open(output_file, "wb") as out_file:
        shutil.copyfileobj(response, out_file)


if __name__ == "__main__":
    generate_data_fetcher_json()



"""Data generation for f-test unit tests."""

import pickle

import numpy as np
from sklearn.model_selection import ParameterGrid

from brainstat.stats.SLM import SLM, f_test
from brainstat.stats.terms import FixedEffect
from brainstat.tests.testutil import datadir


def generate_random_two_slms(I):
    slm1 = SLM(FixedEffect(1), FixedEffect(1))
    slm2 = SLM(FixedEffect(1), FixedEffect(2))

    for key in I.keys():
        if "1" in key:
            setattr(slm1, key[4:], I[key])
        elif "2" in key:
            setattr(slm2, key[4:], I[key])
    return slm1, slm2


def generate_f_test_out(slm1, slm2):
    D = {}
    slm_OUT = f_test(slm1, slm2)
    mykeys = ["X", "df", "SSE", "coef", "k", "t"]
    for key in mykeys:
        D[key] = getattr(slm_OUT, key)
    return D


def params2files(I, D, test_num):
    """Converts params to input/output files"""
    basename = "xstatf"
    fin_name = datadir(basename + "_" + f"{test_num:02d}" + "_IN.pkl")
    fout_name = datadir(basename + "_" + f"{test_num:02d}" + "_OUT.pkl")
    with open(fin_name, "wb") as f:
        pickle.dump(I, f, protocol=4)
    with open(fout_name, "wb") as g:
        pickle.dump(D, g, protocol=4)
    return


def generate_test_data():
    np.random.seed(0)

    a = 40
    b = 30
    c = 100
    d = 42
    e = 77

    mygrid = [
        {
            "slm1X": [np.random.randint(0, a, size=(a, b)), np.random.rand(a, b)],
            "slm1df": [int(a)],
            "slm2df": [int(b)],
            "slm1SSE": [
                np.random.randint(1, a, size=(c, 1)),
                np.random.randint(1, a, size=(c, d)),
            ],
            "slm2SSE": [
                np.random.randint(1, a, size=(c, 1)),
                np.random.randint(1, a, size=(c, d)),
            ],
            "slm1coef": [np.random.rand(e, d), np.random.rand(e, d, 3)],
            "slm2coef": [np.random.rand(e, d)],
        }
    ]

    myparamgrid = ParameterGrid(mygrid)

    # Here wo go!
    # Test 1-16 : slm1X 2D array type int or float64, slm1df and slm2df int,
    # slm1SSE and slm2SSE 2D arrays type int, slm1coef and slm2coef 2D or 3D
    # arrays type float64

    test_num = 0
    for params in myparamgrid:
        test_num += 1
        I = {}
        for key in params.keys():
            I[key] = params[key]
        # make "slm2X" equal to "slm1X" to avoid nested-model error
        I["slm2X"] = I["slm1X"]
        slm1, slm2 = generate_random_two_slms(I)
        D = generate_f_test_out(slm1, slm2)
        params2files(I, D, test_num)


if __name__ == "__main__":
    generate_test_data()



"""Unit tests of peak_clus."""
import pickle

import numpy as np
import pytest

from brainstat.stats._multiple_comparisons import peak_clus
from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect

from .testutil import datadir


def dummy_test(infile, expfile):

    # load input test data
    ifile = open(infile, "br")
    idic = pickle.load(ifile)
    ifile.close()

    slm = SLM(FixedEffect(1), FixedEffect(1))
    slm.t = idic["t"]
    slm.tri = idic["tri"]
    slm.mask = idic["mask"]
    slm.df = idic["df"]
    slm.k = idic["k"]
    slm.resl = idic["resl"]

    thresh = idic["thresh"]
    reselspvert = idic["reselspvert"]
    edg = idic["edg"]

    # call python function
    P_peak, P_clus, P_clusid = peak_clus(slm, thresh, reselspvert, edg)

    # load expected outout data
    efile = open(expfile, "br")
    expdic = pickle.load(efile)
    efile.close()

    O_peak = expdic["peak"]
    O_clus = expdic["clus"]
    O_clusid = expdic["clusid"]

    testout = []

    if isinstance(P_peak, dict):
        for key in P_peak.keys():
            comp = np.allclose(P_peak[key], O_peak[key], rtol=1e-05, equal_nan=True)
            testout.append(comp)
    else:
        comp = np.allclose(P_peak, O_peak, rtol=1e-05, equal_nan=True)

    if isinstance(P_clus, dict):
        for key in P_clus.keys():
            comp = np.allclose(P_clus[key], O_clus[key], rtol=1e-05, equal_nan=True)
    else:
        comp = np.allclose(P_clus, O_clus, rtol=1e-05, equal_nan=True)
    testout.append(comp)

    testout.append(np.allclose(P_clusid, O_clusid, rtol=1e-05, equal_nan=True))

    assert all(flag == True for (flag) in testout)


expected_number_of_tests = 72
parametrize = pytest.mark.parametrize


@parametrize("test_number", range(1, expected_number_of_tests + 1))
def test_run_all(test_number):
    infile = datadir("xstatpeakc_" + f"{test_number:02d}" + "_IN.pkl")
    expfile = datadir("xstatpeakc_" + f"{test_number:02d}" + "_OUT.pkl")
    dummy_test(infile, expfile)



"""Unit tests of mesh_edges."""

import pickle

import numpy as np

from brainstat.mesh.utils import mesh_edges

from .testutil import datadir


def dummy_test(infile, expfile):

    # load input test data
    ifile = open(infile, "br")
    idic = pickle.load(ifile)
    ifile.close()

    surf = {}

    if "tri" in idic.keys():
        surf["tri"] = idic["tri"]

    if "lat" in idic.keys():
        surf["lat"] = idic["lat"]

    # run mesh_edges
    out_edge = mesh_edges(surf)

    # load expected outout data
    efile = open(expfile, "br")
    expdic = pickle.load(efile)
    efile.close()
    exp_edge = expdic["edg"]

    testout = []

    comp = np.allclose(out_edge, exp_edge, rtol=1e-05, equal_nan=True)
    testout.append(comp)

    assert all(flag == True for (flag) in testout)


# data *pkl consists of either keys ['tri'] or ['lat'], which will be assigned to
# the surf{} dictionary while testing
def test_01():
    infile = datadir("xstatedg_01_IN.pkl")
    expfile = datadir("xstatedg_01_OUT.pkl")
    dummy_test(infile, expfile)


def test_02():
    infile = datadir("xstatedg_02_IN.pkl")
    expfile = datadir("xstatedg_02_OUT.pkl")
    dummy_test(infile, expfile)


def test_03():
    infile = datadir("xstatedg_03_IN.pkl")
    expfile = datadir("xstatedg_03_OUT.pkl")
    dummy_test(infile, expfile)


def test_04():
    infile = datadir("xstatedg_04_IN.pkl")
    expfile = datadir("xstatedg_04_OUT.pkl")
    dummy_test(infile, expfile)



"""Unit tests of compute_resels."""
import pickle

import numpy as np
import pytest

from brainstat.stats._multiple_comparisons import compute_resels
from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect
from brainstat.tests.testutil import datadir


def dummy_test(infile, expfile):

    # load input test data
    ifile = open(infile, "br")
    idic = pickle.load(ifile)
    ifile.close()

    slm = SLM(FixedEffect(1), FixedEffect(1))
    for key in idic.keys():
        setattr(slm, key, idic[key])

    resels_py, reselspvert_py, edg_py = compute_resels(slm)

    out = {}
    out["resels"] = resels_py
    out["reselspvert"] = reselspvert_py
    out["edg"] = edg_py

    # load expected outout data
    efile = open(expfile, "br")
    expdic = pickle.load(efile)
    efile.close()

    testout = []

    for key in out.keys():
        if out[key] is not None and expdic[key] is not None:
            comp = np.allclose(out[key], expdic[key], rtol=1e-05, equal_nan=True)
            testout.append(comp)

    assert all(flag == True for (flag) in testout)


expected_number_of_tests = 12
parametrize = pytest.mark.parametrize


@parametrize("test_number", range(1, expected_number_of_tests + 1))
def test_run_all(test_number):
    infile = datadir("xstatresl_" + f"{test_number:02d}" + "_IN.pkl")
    expfile = datadir("xstatresl_" + f"{test_number:02d}" + "_OUT.pkl")
    dummy_test(infile, expfile)



# type: ignore
""" Standard Linear regression models. """
import warnings
from cmath import sqrt
from pathlib import Path
from pprint import pformat
from typing import Any, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from brainspace.mesh.mesh_elements import get_cells, get_points
from brainspace.vtk_interface.wrappers.data_object import BSPolyData

from brainstat._typing import ArrayLike
from brainstat._utils import deprecated
from brainstat.datasets import fetch_parcellation, fetch_template_surface
from brainstat.datasets.base import fetch_template_surface, fetch_yeo_networks_metadata
from brainstat.mesh.utils import _mask_edges, mesh_edges
from brainstat.stats.terms import FixedEffect, MixedEffect
from brainstat.stats.utils import apply_mask, undo_mask


class SLM:
    """Core Class for running BrainStat linear models"""

    # Import class methods
    from ._linear_model import _linear_model
    from ._multiple_comparisons import _fdr, _random_field_theory
    from ._t_test import _t_test

    def __init__(
        self,
        model: Union[FixedEffect, MixedEffect],
        contrast: Union[ArrayLike, FixedEffect],
        surf: Optional[Union[str, dict, BSPolyData]] = None,
        mask: Optional[ArrayLike] = None,
        *,
        correction: Optional[Union[str, Sequence[str]]] = None,
        niter: int = 1,
        thetalim: float = 0.01,
        drlim: float = 0.1,
        two_tailed: bool = True,
        cluster_threshold: float = 0.001,
        data_dir: Optional[Union[str, Path]] = None
    ) -> None:
        """Constructor for the SLM class.

        Parameters
        ----------
        model : brainstat.stats.terms.FixedEffect, brainstat.stats.terms.MixedEffect
            The linear model to be fitted of dimensions (observations, predictors).
        contrast : array-like, brainstat.stats.terms.FixedEffect
            Vector of contrasts in the observations.
        surf : str, dict, BSPolyData, optional
            A surface provided as either a dictionary with keys 'tri' for its
            faces (n-by-3 array) and 'coord' for its coordinates (3-by-n array),
            or as a BrainSpace BSPolyData object, or a string containing a
            template name accepted by fetch_template_surface, by default None.
        mask : array-like, optional
            A mask containing True for vertices to include in the analysis, by
            default None.
        correction : str, Sequence, optional
            String or sequence of strings. If it contains "rft" a random field
            theory multiple comparisons correction will be run. If it contains
            "fdr" a false discovery rate multiple comparisons correction will be
            run. Both may be provided. By default None.
        niter : int, optional
            Number of iterations of the Fisher scoring algorithm for fitting
            mixed effects models, by default 1.
        thetalim : float, optional
            Lower limit on variance coefficients in standard deviations, by
            default 0.01.
        drlim : float, optional
            Step of ratio of variance coefficients in standard deviations, by
            default 0.1.
        two_tailed : bool, optional
            Determines whether to return two-tailed or one-tailed p-values. Note
            that multivariate analyses can only be two-tailed, by default True.
        cluster_threshold : float, optional
            P-value threshold or statistic threshold for defining clusters in
            random field theory, by default 0.001.
        data_dir : str, pathlib.Path, optional
            Path to the location to store BrainStat data files, defaults to
            $HOME_DIR/brainstat_data.
        """
        # Input arguments.
        self.model = model
        self.contrast = contrast

        if isinstance(surf, str):
            self.surf_name = surf
            self.surf = fetch_template_surface(self.surf_name)
        else:
            self.surf_name = None
            self.surf = surf

        self.mask = mask
        self.correction = [correction] if isinstance(correction, str) else correction
        self.niter = niter
        self.thetalim = thetalim
        self.drlim = drlim
        self.two_tailed = two_tailed
        self.cluster_threshold = cluster_threshold
        self.data_dir = data_dir

        # Error check
        if self.surf is None:
            if self.correction is not None and "rft" in self.correction:
                raise ValueError("Random Field Theory corrections require a surface.")

        # We have to initialize fit parameters for our unit tests here.
        # TODO: remove this requirement.
        self._reset_fit_parameters()

    def fit(self, Y: np.ndarray) -> None:
        """Fits the SLM model

        Parameters
        ----------
        Y : numpy.array
            Input data (observation, vertex, variate)

        Raises
        ------
        ValueError
            An error will be thrown when multivariate data is provided and a
            one-tailed test is requested.
        """
        if Y.ndim < 2 or Y.ndim > 3:
            raise ValueError("Input data must be two or three dimensional.")
        elif Y.ndim == 3:
            if (not self.two_tailed) and Y.shape[2] > 1:
                raise NotImplementedError(
                    "One-tailed tests are not implemented for multivariate data."
                )
            student_t_test = Y.shape[2] == 1
        else:
            student_t_test = True

        self._reset_fit_parameters()
        if self.mask is not None:
            Y_masked = apply_mask(Y, self.mask, axis=1)
        else:
            Y_masked = Y.copy()
        self._linear_model(Y_masked)
        self._t_test()
        if self.mask is not None:
            self._unmask()
        if self.correction is not None:
            self.multiple_comparison_corrections(student_t_test)

    def multiple_comparison_corrections(self, student_t_test: bool) -> None:
        """Performs multiple comparisons corrections. If a (one-sided) student-t
        test was run, then make it two-tailed if requested."""
        P1, Q1 = self._run_multiple_comparisons()

        if self.two_tailed and student_t_test:
            self.t = -self.t
            P2, Q2 = self._run_multiple_comparisons()
            self.t = -self.t
            self.P = _merge_rft(P1, P2)
            self.Q = _merge_fdr(Q1, Q2)
        else:
            self.P = P1
            # Make variable type consistent with two one tailed outputs.
            for field in ["peak", "clus", "clusid"]:
                if self.P is not None and field in self.P:
                    if field == "clusid":
                        self.P[field] = [self.P[field]]
                    else:
                        for field2 in self.P[field].keys():
                            self.P[field][field2] = [self.P[field][field2]]
            self.Q = Q1

        self._surfstat_to_brainstat_rft()

    def _run_multiple_comparisons(self) -> Tuple[Optional[dict], Optional[np.ndarray]]:
        """Runs the multiple comparisons tests and returns their outputs.

        Returns
        -------
        dict, None
            Results of random_field_theory. None if not requested.
        np.array, None
            Results of fdr. None if not requested.
        """
        P = None
        Q = None
        if "rft" in self.correction:
            P = {}
            P["pval"], P["peak"], P["clus"], P["clusid"] = self._random_field_theory()
        if "fdr" in self.correction:
            Q = self._fdr()
        return P, Q

    def _reset_fit_parameters(self) -> None:
        """Sets empty parameters before fitting. Prevents issues arising from
        using the same object to fit twice.
        """
        self.X = None
        self.t = None
        self.df = None
        self.SSE = None
        self.coef = None
        self.V = None
        self.k = None
        self.r = None
        self.dr = None
        self.resl = None
        self.tri = None
        self.lat = None
        self.c = None
        self.ef = None
        self.sd = None
        self.dfs = None
        self.P = None
        self.Q = None
        self.du = None

    def _surfstat_to_brainstat_rft(self) -> None:
        """Convert SurfStat RFT output to BrainStat RFT output.
        In short, we convert the peak/clus arrays to pandas dataframes and add Yeo networks.
        """
        if self.P is not None:
            if "peak" in self.P:
                for i in range(len(self.P["peak"]["t"])):
                    if self.P["peak"]["vertid"][i] is not None:
                        self.P["peak"]["vertid"][i] = self.P["peak"]["vertid"][
                            i
                        ].astype(int)
                    if self.P["peak"]["clusid"][i] is not None:
                        self.P["peak"]["clusid"][i] = self.P["peak"]["clusid"][
                            i
                        ].astype(int)

                    if self.surf_name in (
                        "fsaverage",
                        "fsaverage5",
                        "fslr32k",
                        "civet41k",
                        "civet164k",
                    ):
                        yeo7 = fetch_parcellation(
                            self.surf_name, "yeo", 7, data_dir=self.data_dir
                        )
                        yeo_names, _ = fetch_yeo_networks_metadata(7)
                        yeo_names.insert(0, "Undefined")
                        yeo7_index = yeo7[self.P["peak"]["vertid"][i]]
                        if "yeo7" not in self.P["peak"]:
                            self.P["peak"]["yeo7"] = []
                        self.P["peak"]["yeo7"][i] = np.take(yeo_names, yeo7_index)

            if "clus" in self.P:
                for i in range(len(self.P["clus"]["clusid"])):
                    if self.P["clus"]["clusid"][i] is not None:
                        self.P["clus"]["clusid"][i] = self.P["clus"]["clusid"][
                            i
                        ].astype(int)

            none_squeeze = lambda x: np.squeeze(x) if x is not None else None
            for field in ["peak", "clus"]:
                P_tmp = []
                for i in range(len(self.P[field]["P"])):
                    tail_dict = {
                        key: none_squeeze(value[i])
                        for key, value in self.P[field].items()
                    }
                    if tail_dict["P"] is not None:
                        if tail_dict["P"].size == 1:
                            P_tmp.append(pd.DataFrame.from_dict([tail_dict]))
                        else:
                            P_tmp.append(pd.DataFrame.from_dict(tail_dict))
                            P_tmp[i].sort_values(by="P", ascending=True)
                    else:
                        P_tmp.append(pd.DataFrame(columns=tail_dict.keys()))
                self.P[field] = P_tmp

    def _unmask(self) -> None:
        """Changes all masked parameters to their input dimensions."""
        simple_unmask_parameters = ["t", "coef", "SSE", "r", "ef", "sd", "dfs"]
        for key in simple_unmask_parameters:
            attr = getattr(self, key)
            if attr is not None:
                setattr(self, key, undo_mask(attr, self.mask, axis=1))

        # slm.resl unmask
        if self.resl is not None:
            edges = mesh_edges(self.surf)
            _, idx = _mask_edges(edges, self.mask)
            self.resl = undo_mask(self.resl, idx, axis=0)

    def __str__(self) -> str:
        """Returns a string representation of the model."""
        return pformat(
            {key: value for key, value in self.__dict__.items() if not callable(value)}
        )

    """ Property specifications. """

    @property
    def surf(self) -> Union[BSPolyData, dict, None]:
        return self._surf

    @surf.setter
    def surf(self, value):
        self._surf = value
        if self.surf is not None:
            if isinstance(self.surf, BSPolyData):
                self.tri = np.array(get_cells(self.surf)) + 1
                self.coord = np.array(get_points(self.surf)).T
            else:
                if "tri" in value:
                    self.tri = value["tri"]
                    self.coord = value["coord"]
                elif "lat" in value:
                    self.lat = value["lat"]
                    self.coord = value["coord"]

    @surf.deleter
    def surf(self):
        del self._surf

    @property
    def tri(self) -> np.ndarray:
        return self._tri

    @tri.setter
    def tri(self, value):
        if value is not None and np.any(value < 0):
            raise ValueError("Triangle indices must be non-negative.")
        self._tri = value

    @tri.deleter
    def tri(self):
        del self._tri

    @property
    def lat(self) -> np.ndarray:
        return self._lat

    @lat.setter
    def lat(self, value):
        self._lat = value

    @lat.deleter
    def lat(self):
        del self._lat

    @deprecated(
        "Direct usage of this method is deprecated. Please use the `fit()` method instead."
    )
    def t_test(self) -> None:
        raise NotImplementedError

    @deprecated(
        "Direct usage of this method is deprecated. Please use the `fit()` method instead."
    )
    def linear_model(self, Y: Any) -> None:
        raise NotImplementedError

    @deprecated(
        "Direct usage of this method is deprecated. Please use the `fit()` method instead."
    )
    def fdr(self) -> None:
        raise NotImplementedError

    @deprecated(
        "Direct usage of this method is deprecated. Please use the `fit()` method instead."
    )
    def random_field_theory(self) -> None:
        raise NotImplementedError


def _merge_rft(P1: dict, P2: dict) -> dict:
    """Merge two one-tailed outputs of the random_field_theory function.

    Parameters
    ----------
    P1 : dict
        Output dict of random_field_theory
    P2 : dict
        Output dict of random_field_theory

    Returns
    -------
    dict
        Two-tailed version of the inputs.
    """
    if P1 is None and P2 is None:
        return None

    P = {}
    for key1 in P1 or P2:
        P[key1] = {}
        if key1 == "clusid":
            P[key1] = [P1[key1], P2[key1]]
            continue
        for key2 in P1[key1]:
            if key1 == "pval":
                P[key1][key2] = _onetailed_to_twotailed(P1[key1][key2], P2[key1][key2])
            else:
                P[key1][key2] = [P1[key1][key2], P2[key1][key2]]
    return P


def _merge_fdr(Q1: Optional[ArrayLike], Q2: Optional[ArrayLike]) -> np.ndarray:
    """Merge two one-tailed outputs of the fdr function.

    Parameters
    ----------
    Q1 : array-like, None
        Q-values
    Q2 : array-like, None
        Q-values

    Returns
    -------
    np.ndarray
        Two-tailed FDR p-values
    """
    if Q1 is None and Q2 is None:
        return None
    return _onetailed_to_twotailed(Q1, Q2)


def _onetailed_to_twotailed(p1: ArrayLike, p2: ArrayLike) -> np.ndarray:
    """Converts two one-tailed tests to a two-tailed test"""
    if p1 is None and p2 is None:
        return None
    elif p1 is None:
        p1 = p2
    elif p2 is None:
        p2 = p1

    return np.minimum(np.minimum(p1, p2) * 2, 1)


def f_test(slm1: SLM, slm2: SLM) -> SLM:
    """F-statistics for comparing two uni- or multi-variate fixed effects models.

    Parameters
    ----------
    slm1 : brainstat.stats.SLM.SLM
        Standard linear model returned by the t_test function; see Notes for
        details.
    slm2 : brainstat.stats.SLM.SLM
        Standard linear model returned by the t_test function; see Notes for
        details.

    Returns
    -------
    brainstat.stats.SLM.SLM
        Standard linear model with f-test results included.

    """

    if slm1.r is not None or slm2.r is not None:
        warnings.warn("Mixed effects models not programmed yet.")

    slm = SLM(FixedEffect(1), FixedEffect(1))
    if slm1.df > slm2.df:
        X1 = slm1.X
        X2 = slm2.X
        df1 = slm1.df
        df2 = slm2.df
        SSE1 = slm1.SSE
        SSE2 = slm2.SSE
        for key in slm2.__dict__:
            setattr(slm, key, getattr(slm2, key))
    else:
        X1 = slm2.X
        X2 = slm1.X
        df1 = slm2.df
        df2 = slm1.df
        SSE1 = slm2.SSE
        SSE2 = slm1.SSE
        for key in slm1.__dict__:
            setattr(slm, key, getattr(slm1, key))

    r = X1 - np.dot(np.dot(X2, np.linalg.pinv(X2)), X1)
    d = np.sum(r.flatten() ** 2) / np.sum(X1.flatten() ** 2)

    if d > np.spacing(1):
        print("Models are not nested.")
        return

    slm.df = np.array([[df1 - df2, df2]])
    h = SSE1 - SSE2

    # if slm['coef'] is 3D and third dimension is 1, then squeeze it to 2D
    if np.ndim(slm.coef) == 3 and np.shape(slm.coef)[2] == 1:
        x1, x2, x3 = np.shape(slm.coef)
        slm.coef = slm.coef.reshape(x1, x2)

    if np.ndim(slm.coef) == 2:
        slm.k = np.array(1)
        slm.t = np.dot(h / (SSE2 + (SSE2 <= 0)) * (SSE2 > 0), df2 / (df1 - df2))
    elif np.ndim(slm.coef) > 2:
        k2, v = np.shape(SSE2)
        k = np.around((np.sqrt(1 + 8 * k2) - 1) / 2)
        slm.k = np.array(k)
        if k > 3:
            print("Roy's max root for k>3 not programmed yet.")
            return

        l = min(k, df1 - df2)
        slm.t = np.zeros((int(l), int(v)))

        if k == 2:
            det = SSE2[0, :] * SSE2[2, :] - SSE2[1, :] ** 2
            a11 = SSE2[2, :] * h[0, :] - SSE2[1, :] * h[1, :]
            a21 = SSE2[0, :] * h[1, :] - SSE2[1, :] * h[0, :]
            a12 = SSE2[2, :] * h[1, :] - SSE2[1, :] * h[2, :]
            a22 = SSE2[0, :] * h[2, :] - SSE2[1, :] * h[1, :]
            a0 = a11 * a22 - a12 * a21
            a1 = (a11 + a22) / 2
            s1 = np.array([sqrt(x) for x in (a1 ** 2 - a0)]).real
            d = (df2 / (df1 - df2)) / (det + (det <= 0)) * (det > 0)
            slm.t[0, :] = (a1 + s1) * d
            if l == 2:
                slm.t[1, :] = (a1 - s1) * d
        if k == 3:
            det = (
                SSE2[0, :] * (SSE2[2, :] * SSE2[5, :] - SSE2[4, :] ** 2)
                - SSE2[5, :] * SSE2[1, :] ** 2
                + SSE2[3, :] * (SSE2[1, :] * SSE2[4, :] * 2 - SSE2[2, :] * SSE2[3, :])
            )
            m1 = SSE2[2, :] * SSE2[5, :] - SSE2[4, :] ** 2
            m3 = SSE2[0, :] * SSE2[5, :] - SSE2[3, :] ** 2
            m6 = SSE2[0, :] * SSE2[2, :] - SSE2[1, :] ** 2
            m2 = SSE2[3, :] * SSE2[4, :] - SSE2[1, :] * SSE2[5, :]
            m4 = SSE2[1, :] * SSE2[4, :] - SSE2[2, :] * SSE2[3, :]
            m5 = SSE2[1, :] * SSE2[3, :] - SSE2[0, :] * SSE2[4, :]
            a11 = m1 * h[0, :] + m2 * h[1, :] + m4 * h[3, :]
            a12 = m1 * h[1, :] + m2 * h[2, :] + m4 * h[4, :]
            a13 = m1 * h[3, :] + m2 * h[4, :] + m4 * h[5, :]
            a21 = m2 * h[0, :] + m3 * h[1, :] + m5 * h[3, :]
            a22 = m2 * h[1, :] + m3 * h[2, :] + m5 * h[4, :]
            a23 = m2 * h[3, :] + m3 * h[4, :] + m5 * h[5, :]
            a31 = m4 * h[0, :] + m5 * h[1, :] + m6 * h[3, :]
            a32 = m4 * h[1, :] + m5 * h[2, :] + m6 * h[4, :]
            a33 = m4 * h[3, :] + m5 * h[4, :] + m6 * h[5, :]
            a0 = (
                -a11 * (a22 * a33 - a23 * a32)
                + a12 * (a21 * a33 - a23 * a31)
                - a13 * (a21 * a32 - a22 * a31)
            )
            a1 = a22 * a33 - a23 * a32 + a11 * a33 - a13 * a31 + a11 * a22 - a12 * a21
            a2 = -(a11 + a22 + a33)
            q = a1 / 3 - a2 ** 2 / 9
            r = (a1 * a2 - 3 * a0) / 6 - a2 ** 3 / 27
            s1 = (r + [sqrt(x) for x in (q ** 3 + r ** 2)]) ** (1 / 3)
            z = np.zeros((3, v))
            z[0, :] = 2 * s1.real - a2 / 3
            z[1, :] = -s1.real - a2 / 3 + np.sqrt(3) * s1.imag
            z[2, :] = -s1.real - a2 / 3 - np.sqrt(3) * s1.imag

            if not np.count_nonzero(z) == 0:
                z.sort(axis=0)
                z = z[::-1]
            d = df2 / (df1 - df2) / (det + (det <= 0)) * (det > 0)

            for j in range(0, l):
                slm.t[j, :] = z[j, :] * d
    return slm



"""Unit tests for the histology module."""
import pytest
import requests

from brainstat._utils import read_data_fetcher_json

parametrize = pytest.mark.parametrize
json = read_data_fetcher_json()


@parametrize("template", list(json["bigbrain_profiles"].keys()))
def test_urls(template):
    """Tests whether the histology files can be downloaded.

    Parameters
    ----------
    template : list
        Template names.
    """
    r = requests.head(json["bigbrain_profiles"][template]["url"])
    assert r.status_code == 200



"""Data generation for random field theory unit tests.


Tests 1-4 test for variance in the input data types.
Tests 5-20 test input of the optional variables. 
Test 21 tests for nonsense variables in the SLM.

"""

import pickle

import numpy as np
from brainspace.mesh.mesh_elements import get_cells, get_edges, get_points
from nilearn import datasets
from sklearn.model_selection import ParameterGrid

from brainstat.context.utils import read_surface_gz
from brainstat.tests.testutil import datadir, generate_slm, save_slm


def generate_random_slm(surf, n_var=1, dfs=None, mask=None, cluster_threshold=0.001):
    """Generates a valid SLM for a surface.

    Parameters
    ----------
    surf : BSPolyData
        Brain surface.
    n_var : int, optional
        slm.k, by default 1.
    dfs : np.array, None, optional
        Effective degrees of freedom, by default None.
    mask : np.array, optional
        Boolean mask, by default None.
    cluster_threshold : float, optional
        Cluster threshold, by default 0.001.

    Returns
    -------
    brainstat.stats.SLM
        SLM object.
    """
    edges = get_edges(surf)
    vertices = get_points(surf)

    n_vertices = vertices.shape[0]
    n_edges = edges.shape[0]

    slm = generate_slm(
        t=np.random.random_sample((1, n_vertices)),
        df=np.random.randint(2, 100),
        k=n_var,
        resl=np.random.random_sample((n_edges, 1)),
        surf=surf,
        dfs=dfs,
        mask=mask,
        cluster_threshold=cluster_threshold,
    )
    return slm


def slm2files(slm, basename, test_num):
    """Converts an SLM to its input/output files

    Parameters
    ----------
    slm : brainstat.stats.SLM
        SLM object.
    basename : str
        Base name for the file.
    test_num : int
        Number of the test.
    """
    D = {}
    D["pval"], D["peak"], D["clus"], D["clusid"] = slm.random_field_theory()
    save_slm(slm, basename, test_num, input=True)
    filename = datadir(basename + "_" + f"{test_num:02d}" + "_OUT.pkl")
    if "_tri" in D:
        D.pop("_tri")
    with open(filename, "wb") as f:
        pickle.dump(D, f, protocol=4)


def generate_test_data():
    ## Fetch global data and settings.
    # Global test settings
    basename = "xstatp"
    test_num = 1
    np.random.seed(0)

    # Fetch surface data.
    pial_fs5 = datasets.fetch_surf_fsaverage()["pial_left"]
    pial_surf = read_surface_gz(pial_fs5)
    n_vertices = get_points(pial_surf).shape[0]

    ## Define the test parameter grids.
    # Variable types to test
    var_types = {
        "t": [np.float64],
        "df": [int, np.uint16],
        "k": [int, np.uint8],
        "resl": [np.float64],
        "tri": [np.int64],
    }
    type_parameter_grid = ParameterGrid(var_types)

    # Optional variable test.
    var_optional = {
        "dfs": [None, np.random.randint(1, 100, (1, n_vertices))],
        "cluster_threshold": [0.1, 2],
        "mask": [None, np.random.rand(n_vertices) > 0.1],
        "k": [1, 3],
    }

    # Nonsense variables to add.
    var_nonsense = ["X", "coef", "SSE", "c", "ef", "sd"]

    ## Generate test data
    # Variable type tests
    for params in type_parameter_grid:
        slm = generate_random_slm(pial_surf)
        for key in list(params.keys()):
            attr = getattr(slm, key)
            setattr(slm, key, params[key](attr))
        slm2files(slm, basename, test_num)
        test_num += 1

    # Additional variable tests.
    additional_parameter_grid = ParameterGrid(var_optional)
    for params in additional_parameter_grid:
        slm = generate_random_slm(pial_surf)
        for key in list(params.keys()):
            setattr(slm, key, params[key])
        slm2files(slm, basename, test_num)
        test_num += 1

    # Nonsense variable tests.
    slm = generate_random_slm(pial_surf)
    slm.dfs = np.random.randint(1, 100, (1, n_vertices))
    slm.mask = np.random.rand(n_vertices) > 0.1
    for key in var_nonsense:
        if getattr(slm, key) is None:
            setattr(
                slm,
                key,
                np.random.rand(np.random.randint(1, 10), np.random.randint(1, 10)),
            )
    slm2files(slm, basename, test_num)
    test_num += 1


if __name__ == "__main__":
    generate_test_data()



""" Load external datasets. """
import tempfile
import zipfile
from pathlib import Path
from typing import List, Optional, Tuple, Union

import h5py
import numpy as np
from brainspace.mesh.mesh_creation import build_polydata
from brainspace.mesh.mesh_io import read_surface
from brainspace.mesh.mesh_operations import combine_surfaces
from brainspace.vtk_interface.wrappers.data_object import BSPolyData
from netneurotools import datasets as nnt_datasets
from netneurotools.civet import read_civet
from nibabel import load as nib_load
from nibabel.freesurfer.io import read_annot, read_geometry

from brainstat._utils import _download_file, data_directories, read_data_fetcher_json


def fetch_parcellation(
    template: str,
    atlas: str,
    n_regions: int,
    join: bool = True,
    seven_networks: bool = True,
    data_dir: Optional[Union[str, Path]] = None,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Loads the surface parcellation of a given atlas.

    Parameters
    ----------
    template : str,
        The surface template. Valid values are "fsaverage", "fsaverage5",
        "fsaverage6", "fslr32k", by default "fsaverage5".
    atlas : str
        Name of the atlas. Valid names are "cammoun", "glasser", "schaefer", "yeo".
    n_regions : int
        Number of regions of the requested atlas. Valid values for the cammoun
        atlas are 33, 60, 125, 250, 500. Valid values for the glasser atlas are
        360. Valid values for the "schaefer" atlas are 100, 200, 300, 400, 500,
        600, 800, 1000. Valid values for "yeo" are 7 and 17.
    join : bool, optional
        If true, returns parcellation as a single array, if false, returns an
        array per hemisphere, by default True.
    seven_networks : bool, optional
        If true, uses the 7 networks parcellation. Only used for the Schaefer
        atlas, by default True.
    data_dir : str, pathlib.Path, optional
        Directory to save the data, defaults to $HOME_DIR/brainstat_data/parcellation_data.

    Returns
    -------
    np.ndarray or tuple of np.npdarray
        Surface parcellation. If a tuple, then the first element is the left hemisphere.
    """

    data_dir = Path(data_dir) if data_dir else data_directories["PARCELLATION_DATA_DIR"]
    data_dir.mkdir(parents=True, exist_ok=True)

    if atlas == "schaefer":
        parcellations = _fetch_schaefer_parcellation(
            template, n_regions, seven_networks, data_dir
        )
    elif atlas == "cammoun":
        parcellations = _fetch_cammoun_parcellation(template, n_regions, data_dir)
    elif atlas == "glasser":
        parcellations = _fetch_glasser_parcellation(template, data_dir)
    elif atlas == "yeo":
        parcellations = _fetch_yeo_parcellation(template, n_regions, data_dir)
    else:
        raise ValueError(f"Invalid atlas: {atlas}")

    if join:
        return np.concatenate((parcellations[0], parcellations[1]), axis=0)
    else:
        return parcellations[0], parcellations[1]


def fetch_template_surface(
    template: str,
    join: bool = True,
    layer: Optional[str] = None,
    data_dir: Optional[Union[str, Path]] = None,
) -> Union[BSPolyData, Tuple[BSPolyData, BSPolyData]]:
    """Loads surface templates.

    Parameters
    ----------
    template : str
        Name of the surface template. Valid values are "fslr32k", "fsaverage",
        "fsaverage3", "fsaverage4", "fsaverage5", "fsaverage6", "civet41k",
        "civet164k".
    join : bool, optional
        If true, returns surfaces as a single object, if false, returns an
        object per hemisphere, by default True.
    layer : str, optional
        Name of the cortical surface of interest. Valid values are "white",
        "smoothwm", "pial", "inflated", "sphere" for fsaverage surfaces and
        "midthickness", "inflated", "vinflated" for "fslr32k". If None,
        defaults to "pial" or "midthickness", by default None.
    data_dir : str, Path, optional
        Directory to save the data, by default
        $HOME_DIR/brainstat_data/surface_data.

    Returns
    -------
    BSPolyData or tuple of BSPolyData
        Output surface(s). If a tuple, then the first element is the left
        hemisphere.
    """

    data_dir = Path(data_dir) if data_dir else data_directories["SURFACE_DATA_DIR"]
    surface_files = _fetch_template_surface_files(template, layer, data_dir)
    if template[:9] == "fsaverage":
        surfaces_fs = [read_geometry(file) for file in surface_files]
        surfaces = [build_polydata(surface[0], surface[1]) for surface in surfaces_fs]
    elif template == "fslr32k":
        surfaces = [read_surface(file) for file in surface_files]
    else:
        surfaces_obj = [read_civet(file) for file in surface_files]
        surfaces = [build_polydata(surface[0], surface[1]) for surface in surfaces_obj]

    if join:
        return combine_surfaces(surfaces[0], surfaces[1])
    else:
        return surfaces[0], surfaces[1]


def fetch_mask(
    template: str,
    join: bool = True,
    data_dir: Optional[Union[str, Path]] = None,
    overwrite: bool = False,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Fetches midline masks.

    Parameters
    ----------
    template : str
        Name of the surface template. Valid templates are "civet41k" and "civet164k".
    join : bool, optional
        If true, returns a numpy array containing the mask. If false, returns a
        tuple containing the left and right hemispheric masks, respectively, by
        default True
    data_dir : str, pathlib.Path, optional
        Directory to save the data, by default $HOME_DIR/brainstat_data/surface_data.

    Returns
    -------
    numpy.ndarray or tuple of numpy.ndarray
        Midline mask, either as a single array or a tuple of a left and right hemispheric
        array.
    """
    data_dir = Path(data_dir) if data_dir else data_directories["SURFACE_DATA_DIR"]
    mask_file = data_dir / f"{template}_mask.csv"
    url = read_data_fetcher_json()["masks"][template]["url"]
    _download_file(url, mask_file, overwrite=overwrite)

    mask = np.loadtxt(mask_file, delimiter=",") == 1
    if join:
        return mask
    else:
        n = len(mask)
        return mask[: n // 2], mask[n // 2 :]


def fetch_gradients(
    template: str = "fsaverage5",
    name: str = "margulies2016",
    data_dir: Optional[Union[str, Path]] = None,
    overwrite: bool = False,
) -> np.ndarray:
    """Fetch example gradients.

    Parameters
    ----------
    template : str, optional
        Name of the template surface. Valid values are "fsaverage5",
        "fsaverage", "fslr32k", defaults to "fsaverage5".
    name : str
        Name of the gradients. Valid values are "margulies2016", defaults to
        "margulies2016".
    data_dir : str, Path, optional
        Path to the directory to store the gradient data files, by
        default $HOME_DIR/brainstat_data/gradient_data.
    overwrite : bool, optional
        If true, overwrites existing files, by default False.

    Returns
    -------
    numpy.ndarray
        Vertex-by-gradient matrix.
    """
    data_dir = Path(data_dir) if data_dir else data_directories["GRADIENT_DATA_DIR"]
    data_dir.mkdir(parents=True, exist_ok=True)

    gradients_file = data_dir / f"gradients_{name}.h5"
    if not gradients_file.exists() or overwrite:
        url = read_data_fetcher_json()["gradients"][name]["url"]
        _download_file(url, gradients_file, overwrite=overwrite)

    hf = h5py.File(gradients_file, "r")
    return np.array(hf[template]).T


def fetch_yeo_networks_metadata(n: int) -> Tuple[List[str], np.ndarray]:
    """Fetch Yeo networks metadata.

    Parameters
    ----------
    n : int
        Number of Yeo networks, either 7 or 17.

    Returns
    -------
    list of str
        Names of Yeo networks.
    np.ndarray
        Colormap for the Yeo networks.
    """

    if n == 7:
        network_names = [
            "Visual",
            "Somatomotor",
            "Dorsal Attention",
            "Ventral Attention",
            "Limbic",
            "Frontoparietal",
            "Default mode",
        ]
        colormap = (
            np.array(
                [
                    [120, 18, 134],
                    [70, 130, 180],
                    [0, 118, 14],
                    [196, 58, 250],
                    [220, 248, 164],
                    [230, 148, 34],
                    [205, 62, 78],
                ]
            )
            / 255
        )
    elif n == 17:
        network_names = [
            "Visual A",
            "Visual B",
            "Somatomotor A",
            "Somatomotor B",
            "Dorsal Attention A",
            "Dorsal Attention B",
            "Salience / Ventral Attention A",
            "Salience / Ventral Attention B",
            "Limbic A",
            "Limbic B",
            "Frontoparietal C",
            "Frontoparietal A",
            "Frontoparietal B",
            "Temporal Parietal",
            "Default C",
            "Default A",
            "Default B",
        ]
        colormap = (
            np.array(
                [
                    [120, 18, 134],
                    [255, 0, 0],
                    [70, 130, 180],
                    [42, 204, 164],
                    [74, 155, 60],
                    [0, 118, 14],
                    [196, 58, 250],
                    [255, 152, 213],
                    [220, 248, 164],
                    [122, 135, 50],
                    [119, 140, 176],
                    [230, 148, 34],
                    [135, 50, 74],
                    [12, 48, 255],
                    [0, 0, 130],
                    [255, 255, 0],
                    [205, 62, 78],
                ]
            )
            / 255
        )
    else:
        raise ValueError("Invalid number of Yeo networks.")
    return network_names, colormap


def _fetch_template_surface_files(
    template: str,
    layer: Optional[str] = None,
    data_dir: Optional[Union[str, Path]] = None,
) -> Tuple[str, str]:
    """Fetches surface files.

    Parameters
    ----------
    template : str
        Name of the surface template. Valid values are "fslr32k", "fsaverage",
        "fsaverage3", "fsaverage4", "fsaverage5", "fsaverage6", "civet41k",
        "civet164k".
      layer : str, optional
        Name of the cortical surface of interest. Valid values are "white",
        "smoothwm", "pial", "inflated", "sphere" for fsaverage surfaces,
        "midthickness", "inflated", "vinflated" for "fslr32k", and "mid,
        "white" for civet surfaces. If None, defaults to "pial", "midthickness",
        or "mid", by default None.
    data_dir : str, optional
        Directory to save the data, by default None.

    Returns
    -------
    Tuple of str
        Surface files.
    """

    if template == "fslr32k":
        layer = layer if layer else "midthickness"
        bunch = nnt_datasets.fetch_conte69(data_dir=str(data_dir))
    elif template == "civet41k" or template == "civet164k":
        layer = layer if layer else "mid"
        bunch = nnt_datasets.fetch_civet(
            density=template[5:], version="v2", data_dir=str(data_dir)
        )
    else:
        layer = layer if layer else "pial"
        bunch = nnt_datasets.fetch_fsaverage(version=template, data_dir=str(data_dir))
    return bunch[layer]


def _valid_parcellations() -> dict:
    """Returns a dictionary of valid parcellations."""
    return {
        "cammoun": {
            "n_regions": (33, 60, 125, 250, 500),
            "surfaces": ("fsaverage5", "fsaverage6", "fsaverage", "fslr32k"),
        },
        "glasser": {
            "n_regions": (360,),
            "surfaces": ("fsaverage5", "fsaverage", "fslr32k"),
        },
        "schaefer": {
            "n_regions": (100, 200, 300, 400, 500, 600, 800, 1000),
            "surfaces": ("fsaverage5", "fsaverage6", "fsaverage", "fslr32k"),
        },
        "yeo": {
            "n_regions": (7, 17),
            "surfaces": ("fsaverage5", "fsaverage6", "fsaverage", "fslr32k"),
        },
    }


def _fetch_schaefer_parcellation(
    template: str, n_regions: int, seven_networks: int, data_dir: Path
) -> List[np.ndarray]:
    """Fetches Schaefer parcellations."""
    n_networks = 7 if seven_networks else 17
    key = f"{n_regions}Parcels{n_networks}Networks"
    bunch = nnt_datasets.fetch_schaefer2018(version=template, data_dir=str(data_dir))
    if template == "fslr32k":
        cifti = nib_load(bunch[key])
        parcellation_full = np.squeeze(cifti.get_fdata())
        parcellations = [x for x in np.reshape(parcellation_full, (2, -1))]
    else:
        parcellations = [read_annot(file)[0] for file in bunch[key]]
        parcellations[1][parcellations[1] != 0] += n_regions // 2
    return parcellations


def _fetch_cammoun_parcellation(
    template: str, n_regions: int, data_dir: Path
) -> List[np.ndarray]:
    """Fetches Cammoun parcellations."""
    key = f"scale{n_regions:03}"
    bunch = nnt_datasets.fetch_cammoun2012(version=template, data_dir=str(data_dir))
    if template == "fslr32k":
        gifti = [nib_load(file) for file in bunch[key]]
        parcellations = [x.darrays[0].data for x in gifti]
    else:
        parcellations = [read_annot(file)[0] for file in bunch[key]]
    return parcellations


def _fetch_glasser_parcellation(template: str, data_dir: Path) -> List[np.ndarray]:
    """Fetches Glasser parcellation."""
    urls = read_data_fetcher_json()["parcellations"]["glasser"][template]["url"]
    filepaths = []
    for i, hemi in enumerate(("lh", "rh")):
        filename = "_".join(("glasser", "360", template, hemi)) + "label.gii"
        filepaths.append(data_dir / filename)
        _download_file(urls[i], filepaths[i])
    gifti = [nib_load(file) for file in filepaths]
    parcellations = [x.darrays[0].data for x in gifti]
    parcellations[1] = (parcellations[1] + 180) * (parcellations[1] > 0)
    return parcellations


def _fetch_yeo_parcellation(
    template: str, n_regions: int, data_dir: Path
) -> List[np.ndarray]:
    """Fetches Yeo parcellation."""
    filenames = [
        data_dir / f"{template}_{hemi}_yeo{n_regions}.label.gii"
        for hemi in ("lh", "rh")
    ]
    if not all([x.exists() for x in filenames]):
        url = read_data_fetcher_json()["parcellations"]["yeo"]["url"]
        with tempfile.NamedTemporaryFile(suffix=".zip") as f:
            downloaded_file = Path(f.name)
        try:
            _download_file(url, downloaded_file)
            with zipfile.ZipFile(downloaded_file, "r") as zip_ref:
                zip_ref.extractall(data_dir)
        finally:
            downloaded_file.unlink()

    return [nib_load(file).darrays[0].data for file in filenames]



"""Unit tests of random_field_theory."""
import pickle

import numpy as np
import pytest

from brainstat.stats._multiple_comparisons import _random_field_theory
from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect
from brainstat.tests.testutil import datadir


def dummy_test(infile, expfile):

    # load input test data
    ifile = open(infile, "br")
    idic = pickle.load(ifile)
    ifile.close()

    slm = SLM(FixedEffect(1), FixedEffect(1))
    for key in idic.keys():
        if key == "clusthresh":
            slm.cluster_threshold = idic[key]
        else:
            setattr(slm, key, idic[key])

    empirical_output = _random_field_theory(slm)

    # load expected outout data
    efile = open(expfile, "br")
    expdic = pickle.load(efile)
    efile.close()
    expected_output = (expdic["pval"], expdic["peak"], expdic["clus"], expdic["clusid"])

    testout = []
    for (empirical, expected) in zip(empirical_output, expected_output):
        if isinstance(expected, dict):
            for key in expected:
                if key == "mask":
                    continue
                if empirical[key] is None:
                    testout.append(expected[key] is None)
                else:
                    comp = np.allclose(
                        empirical[key], expected[key], rtol=1e-05, equal_nan=True
                    )
                    testout.append(comp)
        else:
            if empirical is None:
                testout.append(expected is None)
            else:
                if len(expected) != 0:
                    comp = np.allclose(empirical, expected, rtol=1e-05, equal_nan=True)
                    testout.append(comp)

    assert all(flag == True for (flag) in testout)


expected_number_of_tests = 21
parametrize = pytest.mark.parametrize


@parametrize("test_number", range(1, expected_number_of_tests + 1))
def test_run_all(test_number):
    infile = datadir("xstatp_" + f"{test_number:02d}" + "_IN.pkl")
    expfile = datadir("xstatp_" + f"{test_number:02d}" + "_OUT.pkl")
    dummy_test(infile, expfile)



"""Unit tests of f-test."""

import pickle

import numpy as np

from brainstat.stats.SLM import SLM, f_test
from brainstat.stats.terms import FixedEffect

from .testutil import datadir


def dummy_test(infile, expfile):

    # load input test data
    ifile = open(infile, "br")
    idic = pickle.load(ifile)
    ifile.close()

    slm1 = SLM(FixedEffect(1), FixedEffect(1))
    slm2 = SLM(FixedEffect(1), FixedEffect(2))
    for key in idic.keys():
        if "1" in key:
            setattr(slm1, key[4:], idic[key])
        elif "2" in key:
            setattr(slm2, key[4:], idic[key])

    # run f test
    outdic = f_test(slm1, slm2)

    # load expected outout data
    efile = open(expfile, "br")
    expdic = pickle.load(efile)
    efile.close()

    testout = []

    for key in expdic.keys():
        comp = np.allclose(
            getattr(outdic, key), expdic[key], rtol=1e-05, equal_nan=True
        )
        testout.append(comp)

    assert all(flag == True for (flag) in testout)


# test data *pkl consists of slm1* and slm2* keys
# slm1* variables will be assigned to slm1 dictionary, and slm2* to the slm2 dict.


def test_01():
    infile = datadir("xstatf_01_IN.pkl")
    expfile = datadir("xstatf_01_OUT.pkl")
    dummy_test(infile, expfile)


def test_02():
    infile = datadir("xstatf_02_IN.pkl")
    expfile = datadir("xstatf_02_OUT.pkl")
    dummy_test(infile, expfile)


def test_03():
    infile = datadir("xstatf_03_IN.pkl")
    expfile = datadir("xstatf_03_OUT.pkl")
    dummy_test(infile, expfile)


def test_04():
    infile = datadir("xstatf_04_IN.pkl")
    expfile = datadir("xstatf_04_OUT.pkl")
    dummy_test(infile, expfile)


def test_05():
    infile = datadir("xstatf_05_IN.pkl")
    expfile = datadir("xstatf_05_OUT.pkl")
    dummy_test(infile, expfile)


def test_06():
    infile = datadir("xstatf_06_IN.pkl")
    expfile = datadir("xstatf_06_OUT.pkl")
    dummy_test(infile, expfile)


def test_07():
    infile = datadir("xstatf_07_IN.pkl")
    expfile = datadir("xstatf_07_OUT.pkl")
    dummy_test(infile, expfile)


def test_08():
    infile = datadir("xstatf_08_IN.pkl")
    expfile = datadir("xstatf_08_OUT.pkl")
    dummy_test(infile, expfile)


def test_09():
    infile = datadir("xstatf_09_IN.pkl")
    expfile = datadir("xstatf_09_OUT.pkl")
    dummy_test(infile, expfile)


def test_10():
    infile = datadir("xstatf_10_IN.pkl")
    expfile = datadir("xstatf_10_OUT.pkl")
    dummy_test(infile, expfile)


def test_11():
    infile = datadir("xstatf_11_IN.pkl")
    expfile = datadir("xstatf_11_OUT.pkl")
    dummy_test(infile, expfile)


def test_12():
    infile = datadir("xstatf_12_IN.pkl")
    expfile = datadir("xstatf_12_OUT.pkl")
    dummy_test(infile, expfile)


def test_13():
    infile = datadir("xstatf_13_IN.pkl")
    expfile = datadir("xstatf_13_OUT.pkl")
    dummy_test(infile, expfile)


def test_14():
    infile = datadir("xstatf_14_IN.pkl")
    expfile = datadir("xstatf_14_OUT.pkl")
    dummy_test(infile, expfile)


def test_15():
    infile = datadir("xstatf_15_IN.pkl")
    expfile = datadir("xstatf_15_OUT.pkl")
    dummy_test(infile, expfile)


def test_16():
    infile = datadir("xstatf_16_IN.pkl")
    expfile = datadir("xstatf_16_OUT.pkl")
    dummy_test(infile, expfile)
